from .config import (
    get_ipfs_tk_mode,
    IpfsTkModes,
    IPFS_REPO_DIR,
    DEF_IPFS_REPO_DIR,
)

from walytis_beta_tools.log import logger
import os

if get_ipfs_tk_mode() == IpfsTkModes.EMBEDDED:
    if not os.path.exists(IPFS_REPO_DIR):
        if IPFS_REPO_DIR == DEF_IPFS_REPO_DIR:
            os.makedirs(IPFS_REPO_DIR)
        else:
            raise Exception(
                "The path specified in the environment variable IPFS_REPO_DIR "
                f"doesn't exist: {IPFS_REPO_DIR}"
            )

    print(f"IPFS repo: {IPFS_REPO_DIR}")
    from ipfs_node import IpfsNode

    ipfs = IpfsNode(IPFS_REPO_DIR)
else:
    from ipfs_remote import IpfsRemote

    ipfs = IpfsRemote("/dns/localhost/tcp/5001/http")
    try:
        ipfs.wait_till_ipfs_is_running(timeout_sec=5)
    except TimeoutError:
        logger.warning("IPFS isn't running. Waiting for IPFS to start...")
        ipfs.wait_till_ipfs_is_running()
        logger.warning("IPFS running now.")



"""The exceptions used in Walytis_Beta API.

See `walytis_beta_tools.exceptions`
"""

from walytis_beta_tools.exceptions import(
    WalytisReplyDecodeError,
    WalytisBugError,
    NoSuchBlockchainError,
    BlockchainAlreadyExistsError,
    NoSuchInvitationError,
    JoinFailureError,
    BlockNotFoundError,
    BlockCreationError,
    InvalidBlockIdError,
    BlockIntegrityError,
    NotSupposedToHappenError,
)




from __future__ import annotations
from collections.abc import Generator

from datetime import datetime
from typing import Generic, Type, TypeVar

from walytis_beta_tools._experimental.ipfs_interface import ipfs

from walytis_beta_tools.block_model import Block, decode_long_id, short_from_long_id
from ..exceptions import (
    BlockIntegrityError,
    BlockNotFoundError,
)


class BlockLazilyLoaded(Block):

    def __init__(self, long_id: bytearray | bytes):
        """Initialise an empty block object, not yet valid on a blockchain."""
        Block.__init__(self)
        if isinstance(long_id, bytes):
            long_id = bytearray(long_id)
        self._long_id: bytearray = long_id

        self._decoded_id = False

    @classmethod
    def from_id(
        cls, long_id: bytearray
    ):
        return cls(long_id=long_id)

    @classmethod
    def from_block(cls, block: Block):
        """Turn a trusted Block object into a BlockLazilyLoaded"""
        if not isinstance(block, Block):
            raise TypeError(
                f"BlockLazilyLoaded.from_block: block must be of type {Block}"
                f", not {type(block)}"
            )

        block.__class__ = cls
        if (
            block._blockchain_version and
            block._ipfs_cid and
            block._creator_id and
            block._creation_time and
            block._topics and
            block._content_length and
            block._content_hash_algorithm and
            block._content_hash and
            block._n_parents and    # won't work for genesis, but it's only one
            block._parents_hash_algorithm and
            block._parents_hash and
            block._parents
        ):
            block._decoded_id = True
        return block

    def _decode_id(self) -> None:
        metadata = decode_long_id(self.long_id)
        self._blockchain_version = metadata["blockchain_version"]
        self._ipfs_cid = metadata["ipfs_cid"]
        self._creator_id = metadata["creator_id"]
        self._creation_time = metadata["creation_time"]
        self._topics = metadata["topics"]
        self._content_length = metadata["content_length"]
        self._content_hash_algorithm = metadata["content_hash_algorithm"]
        self._content_hash = metadata["content_hash"]
        self._n_parents = metadata["n_parents"]
        self._parents_hash_algorithm = metadata["parents_hash_algorithm"]
        self._parents_hash = metadata["parents_hash"]
        self._parents = metadata["parents"]

    def _load_all(self) -> None:
        """Load all this Block's fields that are normally loaded lazily."""
        self.short_id
        self.content
        self.file_data

    def _load_content(self) -> None:
        block_data = ipfs.files.read(self.ipfs_cid)
        content_separator = block_data.index(bytearray([0, 0, 0, 0, 0]))
        content = block_data[content_separator + 5:]

        self._content = content
        if not self.check_integrity(ignore_filedata=True):
            raise BlockIntegrityError(
                "Integrity check for BlockLazilyLoaded failed."
            )

    @property
    def short_id(self) -> bytearray:
        if not self._short_id:
            self._short_id = short_from_long_id(self.long_id)
        return self._short_id

    @property
    def content(self) -> bytearray:
        if not self._content:
            self._load_content()
        return self._content

    @property
    def file_data(self) -> bytearray:
        if not self._file_data:
            self.generate_file_data(check_integrity=True)
        return self._file_data

    @property
    def ipfs_cid(self) -> str:
        if not self._decoded_id:
            self._decode_id()
        return self._ipfs_cid

    @property
    def creator_id(self) -> bytearray:
        if not self._decoded_id:
            self._decode_id()
        return self._creator_id

    @property
    def creation_time(self) -> datetime:
        if not self._decoded_id:
            self._decode_id()
        return self._creation_time

    @property
    def topics(self) -> list[str]:
        if not self._decoded_id:
            self._decode_id()
        return self._topics

    @property
    def parents(self) -> list[bytearray]:
        if not self._decoded_id:
            self._decode_id()
        return self._parents

    @property
    def blockchain_version(self) -> tuple[int, int, int]:
        if not self._decoded_id:
            self._decode_id()
        return self._blockchain_version


# a type variable restricted to subclasses of Block
BlockType = TypeVar('BlockType', bound=Block)


class BlocksList(dict, Generic[BlockType]):
    block_class: Type[BlockType]

    def __init__(self, block_class: Type[BlockType]):
        # Store the class type to instantiate later
        self.block_class = block_class

    @classmethod
    def from_blocks(
        cls: Type['BlocksList[BlockType]'],
        blocks: list[BlockType],
        block_class: Type[BlockType]
    ) -> 'BlocksList[BlockType]':
        if blocks and not isinstance(blocks[0], block_class):
            raise TypeError(
                f"Blocks are of type {type(blocks[0])}, not {block_class}"
            )
        # Use dict.fromkeys() to create the dictionary efficiently
        blocks_dict = dict([
            (bytes(block.long_id), block) for block in blocks
        ])

        # Cast the dictionary to an instance of BlocksList
        # Create an uninitialized instance of the class
        blocks_list = cls.__new__(cls)

        # Manually initialize the dictionary part with the data
        blocks_list.update(blocks_dict)

        # Manually set the block_class
        blocks_list.block_class = block_class

        return blocks_list

    @classmethod
    def from_block_ids(
        cls: Type['BlocksList[BlockType]'],
        block_ids: list[bytes],
        block_class: Type[BlockType]
    ) -> 'BlocksList[BlockType]':

        if block_ids and bytearray([0, 0, 0, 0]) not in bytearray(block_ids[0]):
            print(block_ids[0])
            raise ValueError(
                "It looks like you passed an short ID or invalid ID as a parameter.")
        if block_ids and isinstance(block_ids[0], bytearray):
            block_ids = [bytes(block_id) for block_id in block_ids]

        # Use dict.fromkeys() to create the dictionary efficiently
        blocks_dict = dict.fromkeys(block_ids, None)

        # Cast the dictionary to an instance of BlocksList
        # Create an uninitialized instance of the class
        blocks_list = cls.__new__(cls)

        # Manually initialize the dictionary part with the data
        blocks_list.update(blocks_dict)

        # Manually set the block_class
        blocks_list.block_class = block_class

        return blocks_list

    def add_block_id(self, block_id: bytes | bytearray):
        if bytearray([0, 0, 0, 0]) not in bytearray(block_id):
            raise ValueError(
                "It looks like you passed an short ID or invalid ID as a parameter.")
        if isinstance(block_id, bytearray):
            block_id = bytes(block_id)
        self.update({block_id: None})

    def add_block(self, block: BlockType):
        if not isinstance(block, self.block_class):
            raise ValueError(
                f"block must be of type {self.block_class}, not {type(block)}"
            )
        dict.__setitem__(self, bytes(block.long_id), block)

    def __getitem__(self,  block_id: bytes) -> BlockType:
        try:
            block = dict.__getitem__(self, block_id)
        except KeyError:
            if bytearray([0, 0, 0, 0]) not in block_id:
                raise ValueError(
                    "It looks like you passed an short ID or invalid ID as a parameter.")
            else:
                raise BlockNotFoundError("Block ID not found in BlocksList")
        if not block:
            block = self.block_class.from_id(bytearray(block_id))
            dict.__setitem__(self, block_id, block)
        return block

    def get_block(self, block_id: bytes) -> BlockType:
        return self.__getitem__(block_id)

    def get_blocks(
        self, reverse: bool = False
    ) -> Generator[BlockType, None, None]:
        block_ids = self.get_long_ids()

        # reverse order if desired
        if reverse:
            block_ids = reversed(block_ids)

        for block_id in block_ids:
            yield self.__getitem__(block_id)

    def get_long_ids(self):
        return list(self.keys())

    def get_short_ids(self):
        return [short_from_long_id(long_id) for long_id in list(self.keys())]

    def get_num_blocks(self) -> int:
        return len(self)



"""The machinery for communicating with Brenthy Core.

It contains functions that are a direct part of Walytis_Beta API and intended
for use by the Walytis_Beta API user (e.g. the create_blockchain() function),
but also functions that are used only by higher-level WalytisAPI constructs
such as BlocksListener class, which is not intended for use by the Walytis API
user, but is used by the blockchain_model.Blockchain class instead,
which is in turn intended for use by the Walytis_Beta API user.
"""

from enum import Enum
import os
from .walytis_beta_generic_interface import BaseWalytisBetaInterface
from datetime import datetime

from walytis_beta_tools.log import logger_api as logger

from walytis_beta_tools.block_model import Block

from walytis_beta_tools.exceptions import (
    BlockchainAlreadyExistsError,  # noqa
    BlockIntegrityError,  # noqa
    BlockNotFoundError,  # noqa
    JoinFailureError,  # noqa
    NoSuchBlockchainError,  # noqa
    NoSuchInvitationError,  # noqa
    WalytisBugError,  # noqa
    WalytisReplyDecodeError,  # noqa
)
from walytis_beta_tools.exceptions import (
    NO_SUCH_BLOCKCHAIN_MESSAGE,
    NO_SUCH_INVITATION_MESSAGE,
    BLOCK_NOT_FOUND,
    WALYTIS_BETA_ERROR_MESSAGE,
    BLOCKCHAIN_EXISTS_MESSAGE,
)
# storing this blockchain's name so that we don't missspell it
WALYTIS_BETA = "Walytis_Beta"


# --------------Settings ---------------------------
# logger.PRINT_DEBUG = False




_waly: BaseWalytisBetaInterface


from walytis_beta_tools._experimental.config import WalytisBetaApiTypes, get_walytis_beta_api_type

match get_walytis_beta_api_type():
    case WalytisBetaApiTypes.WALYTIS_BETA_BRENTHY_API:
        from .walytis_beta_brenthy_api import WalytisBetaNetApi
        _waly = WalytisBetaNetApi()
    case WalytisBetaApiTypes.WALYTIS_BETA_DIRECT_API:
        from .walytis_beta_embedded_api import WalytisBetaDirectInterface
        _waly = WalytisBetaDirectInterface()
    case _:
        raise WalytisBugError(
            "walytis_beta_interface: Didn't take into account how to process "
            f"WALYTIS_BETA_API_TYPE {get_walytis_beta_api_type()}"
        )
logger.info(_waly)
BlocksListener = _waly.BlocksListener


# --- Blockchain IDs and Names ---


def list_blockchains(names_first: bool = False) -> list[tuple[str, str]]:
    """Get the IDs names of all the blockchains we have.

    Args:
        names_first: determines the order of the the tuple contents in the
                        return value. See Returns.

    Returns:
        list: A list of tuples of the names and IDs of all the Walytis_Beta
                blockchains we have. Tuples are (id, name) by default,
                but (name, id) if names_first == True
    """
    return _waly.list_blockchains(names_first=names_first)


def list_blockchain_names() -> list[str]:
    """Get the names of all the blockchains we have.

    Returns:
        list: A list of the names of all the Walytis_Beta blockchains we have.
    """
    return _waly.list_blockchain_names()


def list_blockchain_ids() -> list[str]:
    """Get the ids of all the blockchains we have.

    Returns:
        list: A list of the names of all the Walytis_Beta blockchains we have.
    """
    return _waly.list_blockchain_ids()


def get_blockchain_id(blockchain_name: str | bytearray) -> str:
    """Given a blockchain's name, returns its ID.

    If a blockchain's ID is given, it returns the ID.

    Args:
        blockchain_name (str): the name of the blockchain to look up
    Returns:
        str: the ID of the blockchain
    """
    return _waly.get_blockchain_id(blockchain_name)


def get_blockchain_name(blockchain_id: str) -> str:
    """Given a blockchain's id, returns its name.

    If a blockchain's name is given, it returns the name.

    Args:
        blockchain_id (str): the id of the blockchain to look up
    Returns:
        str: the name of the blockchain
    """
    return _waly.get_blockchain_name(blockchain_id)


# --- Blockchain Lifecycle ---


def create_blockchain(blockchain_name: str = "") -> str:
    """Create and run a new blockchain.

    Args:
        blockchain_name (str): blockchain's local name: a human-readable label
                to ease recognition when manually interacting with Walytis_API.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.

    Returns:
        str: ID of the created blockchain
    """
    return _waly.create_blockchain(blockchain_name)


def join_blockchain(invitation: str, blockchain_name: str = "") -> str:
    """Join a blockchain using an invitation generated by a blockchain node.

    Args:
        invitation (str): the invitation generated by a blockchain member
        blockchain_name (str): blockchain's local name: a human-readable label
                to ease recognition when manually interacting with WalytisAPI.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.

    Returns:
        str: the ID of the joined blockchain
    """
    return _waly.join_blockchain(invitation, blockchain_name)


def join_blockchain_from_zip(
    blockchain_id: str, blockchain_data_path: str, blockchain_name: str = ""
) -> None:
    """Join an existing live blockchain, given a zip file of all its data.

    Args:
        blockchain_id (str): the name of the blockchain
        blockchain_data_path (str): path of the zip file containing the
                blockchain's data
        blockchain_name (str): blockchain's local name: a human-readable label
                to ease recognition when manually interacting with WalytisAPI.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.
    """
    return _waly.join_blockchain_from_zip(
        blockchain_id, blockchain_data_path, blockchain_name
    )


def join_blockchain_from_cid(
    blockchain_id: str, blockchain_data_cid: str, blockchain_name: str = ""
) -> None:
    """Join an existing live blockchain, given the IPFS CID of its appdata.

    Args:
        blockchain_id (str): the name of the blockchain
        blockchain_data_cid (str): IPFS CID of the zip file containing the
                blockchain's data
        blockchain_name (str): blockchain's local name: a human-readable label
                to ease recognition when manually interacting with WalytisAPI.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.
    """
    return _waly.join_blockchain_from_cid(
        blockchain_id, blockchain_data_cid, blockchain_name
    )


def delete_blockchain(blockchain_id: str) -> None:
    """Delete the blockchain with the given id or name.

    Args:
        blockchain_id (str): the id or name of the blockchain to delete
    """
    return _waly.delete_blockchain(blockchain_id)


def create_block(
    blockchain_id: str,
    content: bytearray | bytes,
    topics: list[str] | str | None = None
) -> Block:
    """Create a block with the given content on the specified blockchain.

    Args:
        blockchain_id (str): the id or name of the blockchain to create a block
        content (bytearray): the content to store on the blockchain
        topics (list): user's metadata to encode in the block's ID
    Returns:
        Block: the newly created block
    """
    return _waly.create_block(
        blockchain_id,
        content,
        topics,
    )


def get_block(blockchain_id: str, block_id: bytearray) -> Block:
    """Get a block from the specified blokchain given its block ID.

    Args:
        blockchain_id (str): the id or name of the blockchain from which to get
                                the block
        block_id (str): the id of the block
    Returns:
        Block: the Block object for the requested block
    """
    return _waly.get_block(blockchain_id, block_id)


def is_block_known(blockchain_id: str, block_id: bytearray) -> bool:
    """Check whether or not the specified blockchain has the given block.

    Args:
        blockchain_id (str): the id or name of the blockchain to which the
                                block to look up belongs
        block_id (bytearray): the block ID to look up in the block ID records
    Returns:
        bool: Whether the block is a known block
    """
    return _waly.is_block_known(blockchain_id, block_id)


def get_latest_blocks(
    blockchain_id: str,
    amount: int | None = None,
    since_date: datetime | None = None,
    topics: list[str] | None = None,
    long_ids: bool = False,  # defaults to False for backwards-compatibility
) -> list[bytearray]:
    """Get the latest few blocks on the blockchain.

    The amount of which is determined by the paramters.
    The blocks are returned in order of reception or creation date (CHECK!)
    from most recent to oldest.
    One of the paramaters amount or since_date must be set.

    Args:
        blockchain_id (str): the id or name of the blockchain to look up
        amount (int): how many blocks should be returned
        since_date (datetime): the maximum age of the blocks to look for.
                This function will return several even older blocks,
                some of which may have been received after the specified date
        topics (list): the topics to filter by, returning only
                blocks published on those topics, if this paramater is set.
                If not set, this function returns all blocks within the range
                of the first two parameters

    Returns:
        list: a list of IDs of the latest few blocks
    """
    return _waly.get_latest_blocks(
        blockchain_id=blockchain_id,
        amount=amount,
        since_date=since_date,
        topics=topics,
        long_ids=long_ids,
    )


def get_blockchain_data(blockchain_id: str) -> str:
    """Create a zip file of all this blockchain's data, returning its path.

    Args:
        blockchain_id (str): the id or name of the blockchain to zip
    Returns:
        str: the path of the zip file created
    """
    return _waly. get_blockchain_data(blockchain_id)


def get_peers(blockchain_id: str) -> list[str]:
    """Get a list of IPFS IDs of the given blockchain's currently online nodes.

    Args:
        blockchain_id (str): the id or name of the blockchain to look up
    Returns:
        list[str]: the IPFS IDs of the blockchain's currently online nodes
    """
    return _waly.get_peers(blockchain_id)


def create_invitation(
    blockchain_id: str, one_time: bool = True, shared: bool = False
) -> str:
    """Create a code which other nodes can use to join this blockchain.

    Args:
        blockchain_id (str): the id or name of the blockchain to create a
                        invitation for
        one_time (bool): (optional, default True) whether this code is valid
                        only for one use only instead of being reusable
        shared (bool): whether this invitation should be hosted by other
                        nodes who join this blockchain too, or just by this one
    Returns:
        str: an invitation code that can be used by other computers to join
                this blockchain
    """
    return _waly.create_invitation(
        blockchain_id=blockchain_id, one_time=one_time, shared=shared
    )


def get_invitations(blockchain_id: str) -> list[str]:
    """Get the specified blockchain's active invitations.

    Args:
        blockchain_id (str): the id or name of the blockchain whose invitations
                                to look up
    Returns:
        list: the invitations of this blockchain
    """
    return _waly.get_invitations(blockchain_id)


def delete_invitation(blockchain_id: str, invitation: str) -> None:
    """Delete an invitation from the specified blockchain.

    Args:
        blockchain_id (str): the id or name of the blockchain to delete an
                            invitation from
        invitation (str): the invitation to delete
    """
    return _waly.delete_invitation(blockchain_id, invitation)


def get_walytis_beta_version() -> tuple[int, int, int]:
    """Get the software version of the locally running Walytis node.

    Returns:
        tuple: the software version of the locally running Walytis_Beta node
    """
    return _waly.get_walytis_beta_version()


def get_walytis_beta_version_string() -> str:
    """Get the software version of the locally running Walytis_Beta node.

    Returns:
        str: the software version of the locally running Walytis_Beta node
    """
    return _waly.get_walytis_beta_version_string()


def get_walytis_beta_api_version() -> tuple[int, int, int]:
    """Get the software version of the this walytis_beta_api library.

    Returns:
        tuple: the software version of the this walytis_beta_api library
    """
    return _waly.get_walytis_beta_api_version()


def get_walytis_beta_api_version_string() -> str:
    """Get the software version of the this walytis_beta_api library.

    Returns:
        str: the software version of the this walytis_beta_api library
    """
    return _waly.get_walytis_beta_api_version_string()


def get_and_read_block(short_id: bytearray) -> Block:
    """Given a block's ID, download and verify it.

    Args:
        short_id (str): the ID of the block to get
    Returns:
        Block: an object representing the specified block, containing all its
                metadata and block content
    """
    return _waly.get_and_read_block(short_id)


def read_block(block_data: bytearray, ipfs_cid: str) -> Block:
    """Read a block-file, generating a block object if it's valid.

    Args:
        block_data (bytearray): the raw data of the block to read.
        ipfs_cid (str): the IPFS content ID of the block
    Returns:
        Block: a Block object
    """
    return _waly.read_block(block_data, ipfs_cid)



"""`walytis_beta_api` Blockchain class.

The herein contained Blockchain class isn't the code running the Walytis
blockchains themselves, it is merely a construct in `walytis_beta_api` for
providing the means to interact with a Walytis blockchain.

The real blockchain class for running blockchains in Walytis-Core is in:
Brenthy/blockchains/Walytis_Beta/walytis_beta.py
"""

import os
import shutil
from threading import Lock, Thread
from typing import Callable

import appdirs
from walytis_beta_tools.log import logger_blockchain_model as logger
from brenthy_tools_beta.utils import (
    bytes_to_string,
    function_name,
    string_to_bytes,
)

from walytis_beta_tools._experimental.block_lazy_loading import (
    BlocksList,
    BlockLazilyLoaded,
)
from walytis_beta_tools.block_model import Block, short_from_long_id
from walytis_beta_tools.exceptions import (  # pylint: disable=unused-import
    BlockCreationError,
    BlockNotFoundError,
    NotSupposedToHappenError,
)
from ._experimental.generic_blockchain import (
    _GenericBlockchainImpl as GenericBlockchain,
)
from .walytis_beta_interface import (
    WALYTIS_BETA,
    BlocksListener,
    create_block,
    create_blockchain,
    create_invitation,
    delete_blockchain,
    delete_invitation,
    get_block,
    get_blockchain_data,
    get_blockchain_id,
    get_blockchain_name,
    get_invitations,
    get_latest_blocks,
    get_peers,
    join_blockchain,
)
from collections.abc import Generator
from walytis_beta_tools._experimental.generic_block import GenericBlock
from time import sleep

N_STARTUP_BLOCKS = (
    400  # how many of this blockchain's  blocks we should load on startup
)
# logger.set_print_level("info")


walytis_beta_appdata_dir = os.path.join(appdirs.user_data_dir(), "WalytisApps")


def set_appdata_dir(appdata_dir: str) -> None:
    """Set the appdata directory for keeping track of apps' Walytis usage."""
    global walytis_beta_appdata_dir
    walytis_beta_appdata_dir = appdata_dir
    if not os.path.exists(walytis_beta_appdata_dir):
        os.makedirs(walytis_beta_appdata_dir)


def get_appdata_dir() -> str:
    """Get the appdata directory for keeping track of apps' Walytis usage."""
    return walytis_beta_appdata_dir


# backward-compatibility with old path:
_old_appdata_path = os.path.join(appdirs.user_data_dir(), "BrenthyApps")
if os.path.exists(_old_appdata_path):
    walytis_beta_appdata_dir = _old_appdata_path

# backkward-compatibility with old function name


def get_walytis_appdata_dir() -> str:
    """Get the appdata directory for keeping track of apps' Walytis usage.

    Exactly the same as get_appdata_dir(), kept for backward-compatibility.
    """
    print("DEPRECATING: please use `get_appdata_dir()` instead.")
    return get_appdata_dir()


class Blockchain(GenericBlockchain):
    """Represents a running Walytis_Beta blockchain.

    Provides the means to interact with a Walytis blockchain.
    """

    def __init__(
        self,
        blockchain_id: str,
        block_received_handler: Callable[[Block], None] | None = None,
        app_name: str = "",
        appdata_dir: str = "",
        auto_load_missed_blocks: bool = True,
        forget_appdata: bool = False,
        sequential_block_handling: bool = True,
        update_blockids_before_handling: bool = False,
    ):
        """Create an object to represent a Walytis blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain
            block_received_handler (Callable): function to be called every
                time a new block is received on this blockchain.
                This function should not have a long execution time because
                this class waits for its execution to finish before saving
                the received block in its list of processed blocks.
                If this eventhandler raises an exception, it will be called
                with the same block at a later stage.
                The same block is never passed to this handler twice, and
                no child block will ever be passed to this function before
                its parent block.
            app_name (str): the unique name of this application, to distinguish
                between different applications that use the same blockchain
                for appdata management
            appdata_dir (str): the directory in which to save data on which
                blocks this application has processed for this blockchain
            auto_load_missed_blocks (bool): whether or not to automatically get
                and process the blocks received by the blockchain while
                this application was offline.
            forget_appdata (bool): whether or not to ignore and overwrite any
                existing records of which blocks applications with the
                provided `app_name` have processed
            sequential_block_handling (bool): if True, the provided block-
                received handler will be executed on the main thread, and
                the handler will only be executed for the next block
                after the handler for the last block executes witout
                error.
                If set to False, the block-received handler will always be
                started on a new thread, meaning it is possible for
                multiple calls of the handler to be running in parallel
                for different blocks.
            update_blockids_before_handling (bool): whether or not the
                `blocks` attribute should be updated before running
                `block_received_handler` when a new block is received
        """
        # declare attribute already to avoid error in destructor
        # if error occurs in constructor
        self._blocks_listener = None
        self._blocks = BlocksList(BlockLazilyLoaded)

        self.blockchain_id = get_blockchain_id(blockchain_id)
        self.name = get_blockchain_name(self.blockchain_id)
        self.app_name = app_name
        self._block_received_handler = block_received_handler
        self._block_received_handler_lock = Lock()

        if not isinstance(blockchain_id, str):
            error = TypeError(
                "The blockchain_id parameter must be of type string."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if appdata_dir and app_name:
            error = ValueError(
                "Don't pass both app_name and appdata_dir. "
                "You can pass one of them or none."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

        if not isinstance(app_name, str):
            error = TypeError("The app_name parameter must be of type string.")
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if not isinstance(appdata_dir, str):
            error = TypeError(
                "The appdata_dir parameter must be of type string."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if appdata_dir and not os.path.isdir(appdata_dir):
            _error = FileNotFoundError(
                "The appdata_dir parameter must be an exising directory."
            )
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

        self._blocklist_lock = Lock()
        self._terminate = False

        self.sequential_block_handling = sequential_block_handling
        self.update_blockids_before_handling = update_blockids_before_handling

        self._blocks_listener = BlocksListener(
            self.blockchain_id, self._on_new_block_received
        )

        # if the user specified an appdata directory, use it
        if appdata_dir:
            self.appdata_dir = appdata_dir
        elif app_name:
            self.appdata_dir = os.path.join(
                walytis_beta_appdata_dir,
                self.blockchain_id,
                "Apps",
                self.app_name,
            )
            # if appdata path doesn't exist
            # and we're not being used by a temporary app
            if not os.path.exists(self.appdata_dir):
                os.makedirs(self.appdata_dir)
        else:
            self.appdata_dir = ""
        if forget_appdata and os.path.exists(self.appdata_dir):
            shutil.rmtree(self.appdata_dir)
        self._load_blocks_list()
        if auto_load_missed_blocks:
            self.load_missed_blocks(N_STARTUP_BLOCKS)

    def add_block(
        self,
        content: bytearray | bytes,
        topics: list[str] | str | None = None,
        wait_until_handled: bool = True,
    ) -> Block:
        """Add a new block to this blockchain.

        Args:
            content (bytearay): the content which the new block should have
            topics (list<str>): a list of topics this block belongs to
                    Topics allow blocks to be categorised by their application-
                    level function given the block ID only, without reading the
                    block content
        Returns:
            Block block: the newly created block
        """
        if type(content) not in [bytearray, bytes]:
            raise TypeError(
                "Content must be of type bytes or bytearray, not "
                f"{type(content)}"
            )
        self._blocklist_lock.acquire()

        if topics == []:
            topics = self.blockchain_id
        try:
            block = create_block(self.blockchain_id, content, topics=topics)
        except BlockCreationError as error:
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            self._blocklist_lock.release()
            raise error

        # self._blocks.add_block_id(block.long_id)
        # self._save_blocks_list(True)
        self._blocklist_lock.release()
        if wait_until_handled:
            self._on_new_block_received(block)

            # wait for block received handler to finish for consistent behaviour
            while block.long_id not in self.get_block_ids():
                sleep(0.1)
                if self._terminate:
                    self._on_new_block_received(block, override_terminate=True)
                    break

        return block

    def get_blocks(
        self, reverse: bool = False
    ) -> Generator[BlockLazilyLoaded]:
        return self._blocks.get_blocks(reverse=reverse)

    def get_block_ids(self) -> list[bytes]:
        return self._blocks.get_long_ids()

    def get_num_blocks(self) -> int:
        return len(self._blocks)

    def has_block_id(self, block_id: bytes | bytearray) -> bool:
        if isinstance(block_id, bytearray):
            block_id = bytes(block_id)
        return block_id in self._blocks

    def has_block(self, block: GenericBlock) -> bool:
        return self.has_block_id(block.long_id)

    def get_block(self, id: bytearray | bytes) -> Block:
        """Get a block object from this blockchain given its block ID.

        Args:
            id (bytearray): the id of the block to look up
                            or if int: the index of the block to look up
        Returns:
            Block: a Block object containing all the requested block's data
        """
        self._blocklist_lock.acquire()

        # if index is passed instead of block_id, get block_id from index
        if isinstance(id, int):
            try:
                id = self.get_block_ids()[id]
            except IndexError:
                self._blocklist_lock.release()
                message = (
                    "Walytis_BetaAPI.Blockchain: Get Block from index: "
                    "Index out of range."
                )
                logger.error(message)
                raise IndexError(message)
        else:
            id_bytearray = bytearray(id)
            len_id = len(id_bytearray)
            if (
                bytearray([0, 0, 0, 0]) not in id_bytearray
            ):  # if a short ID was passed
                short_id = None
                for long_id in self.get_block_ids():
                    if bytearray(long_id)[:len_id] == id_bytearray:
                        short_id = long_id
                        break
                if not short_id:
                    raise BlockNotFoundError()
                id = bytes(short_id)
        if isinstance(id, bytearray):
            id = bytes(id)
        try:
            block = self._blocks[id]
            self._blocklist_lock.release()
            return block
        except KeyError:
            self._blocklist_lock.release()

            error = BlockNotFoundError(
                "This block isn't recorded (by brenthy_api.Blockchain) as being "
                "part of this blockchain."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

    def create_invitation(
        self, one_time: bool = True, shared: bool = False
    ) -> str:
        """Create a code which another node can use to join this blockchain.

        Args:
            one_time (bool): whether this code is non-reusable
            shared (bool): whether this invitation should be hosted by other
                    nodes who join this blockchain too, or just by this one
        Returns:
            str: a code that can be used by other nodes to join this blockchain
        """
        return create_invitation(
            self.blockchain_id, one_time=one_time, shared=shared
        )

    def delete_invitation(self, invitation: str) -> None:
        """Delete an invitation from this blockchain.

        Args:
            invitation (str): the invitation to delete
        """
        delete_invitation(self.blockchain_id, invitation)

    def get_invitations(self) -> list[str]:
        """Get a list of the invitations we're hositng for this blockchain.

        Returns:
            list: the invitations we're hosting for this blockchain,
                    as JSON strings
        """
        return get_invitations(self.blockchain_id)

    def get_peers(self) -> list[str]:
        """Get the IPFS peer IDs of blockchain nodes who are currently online.

        Returns:
            list[str]: a list of IPFS peer IDs of the currently connected nodes
        """
        return get_peers(self.blockchain_id)

    def get_blockchain_data(self) -> str:
        """Create a zip file of all this blockchain's data, returning its path.

        Returns:
            str: the path of the zip file created
        """
        return get_blockchain_data(self.blockchain_id)

    def _run_block_received_handler(self, block: Block) -> None:
        if not self.block_received_handler:
            return
        return_value = self.block_received_handler(block)
        if return_value is not None:
            raise ValueError(
                "block_received_handler must return None, as it's meaningful "
                "usage is reserved for future use"
            )

    def _on_new_block_received(
        self,
        block: Block,
        save_blocks_list: bool = True,
        already_locked: bool = False,
        override_terminate: bool = False,
        already_locked_block_received: bool = False,
    ) -> None:
        """Handle a newly received block.

        Gets called when this blockchain received a new block. Saves the block
        in our block list and calls the appropriate user-defined eventhandlers.

        Args:
            block (Block): the newly received block
            save_blocks_list (bool): whether or not to save our updated list
                        of blocks to appdata
            already_locked (bool): whether or not we've already acquired the
                        lock for working with the blocks list
            override_terminate (bool): whether to process block even if this
                        Blockchain object is terminating
        """
        if self._terminate and not override_terminate:
            return
        if not already_locked_block_received:
            self._block_received_handler_lock.acquire()
        # if not already_locked:
        #     self._blocklist_lock.acquire()
        short_ids = self._blocks.get_short_ids()
        if bytes(block.long_id) not in self._blocks:
            for parent in block.parents:
                if bytes(parent) not in short_ids:
                    self._on_new_block_received(
                        get_block(self.blockchain_id, parent),
                        save_blocks_list=save_blocks_list,
                        # already_locked=True
                        already_locked=already_locked,
                        already_locked_block_received=True,
                    )
            try:
                if self.update_blockids_before_handling:
                    self._update_blocks_list(block, already_locked=True)

                # call user's block_received_handler
                # only if the block isn't a genesis block
                if block.topics == ["genesis"]:
                    logger.info("WAPI: Received Genesis Block!")
                elif self.block_received_handler:  # it's a normal block
                    if self.sequential_block_handling:
                        self._run_block_received_handler(block)
                    else:
                        Thread(
                            target=self._run_block_received_handler,
                            args=(block,),
                            name=f"WAPI-{self.name}-block_received_handler",
                        ).start()
                if not self.update_blockids_before_handling:
                    self._update_blocks_list(block, already_locked=True)

            except Exception as e:
                logger.error(f"WAPI: Blockchain._on_new_block_received: {e}")
        if not already_locked_block_received:
            self._block_received_handler_lock.release()

    def _update_blocks_list(
        self,
        block: Block,
        save_blocks_list: bool = True,
        already_locked: bool = False,
    ) -> None:
        if not already_locked:
            self._blocklist_lock.acquire()
        self._blocks.add_block_id(block.long_id)
        if save_blocks_list:
            self._save_blocks_list(already_locked=True)

        if not already_locked:
            self._blocklist_lock.release()

    def _save_blocks_list(self, already_locked: bool = False) -> None:
        """Save our list of known/processed block IDs to an appdata file."""
        # don't save any appdata if we're being used by a temporary app
        if not self.appdata_dir:
            return
        if not already_locked:
            self._blocklist_lock.acquire()

        with open(
            os.path.join(self.appdata_dir, "blocks_list"), "w+"
        ) as filewriter:
            filewriter.writelines(
                [
                    bytes_to_string(block_id) + "\n"
                    for block_id in self.get_block_ids()
                ]
            )
        if not already_locked:
            self._blocklist_lock.release()

    def load_missed_blocks(self, amount: int) -> None:
        """Get recent blocks.

        Ask the Blockchain for recent blocks on this topic that might have been
        received while this program was offline. Any of these missed blocks
        will trigger the self.block_received_handler event handler.

        Args:
            amount (int): the number of blocks to load
        """
        # self._blocklist_lock.acquire()
        logger.info(amount)
        latest_blocks = get_latest_blocks(
            self.blockchain_id, amount=amount, long_ids=True
        )
        if not latest_blocks:
            # self._blocklist_lock.release()
            error = NotSupposedToHappenError(
                "Got no latest blocks from blockchain. "
                "This is probably due to a bug in walytis_api."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        count = 0
        for block_id in latest_blocks:
            if self._terminate:
                # self._blocklist_lock.release()
                return

            if bytes(block_id) not in self.get_block_ids():
                try:
                    block = get_block(self.blockchain_id, block_id)
                    self._on_new_block_received(
                        block,
                        save_blocks_list=False,
                        # already_locked=True
                    )
                except Exception as e:
                    logger.error(f" Blockchain.load_missed_blocks: {e}")
            else:
                count += 1

        self._save_blocks_list(
            # already_locked=True
        )
        # self._blocklist_lock.release()

    def _load_blocks_list(self) -> None:
        """Load our list of known/processed block IDs from an appdata file."""
        if not self.appdata_dir:
            return

        self._blocklist_lock.acquire()

        if os.path.exists(os.path.join(self.appdata_dir, "blocks_list")):
            block_ids = []
            with open(
                os.path.join(self.appdata_dir, "blocks_list"), "r"
            ) as filereader:
                for line in filereader.readlines():
                    block_ids.append(string_to_bytes(line[:-1]))

            # AppData migration from storing short IDs to long_ids
            # for BACKWARDS COMPATIBILITY
            if bytearray([0, 0, 0, 0]) not in bytearray(
                block_ids[0]
            ) or bytearray([0, 0, 0, 0]) not in bytearray(block_ids[-1]):
                print("Migrating...")
                block_long_ids = dict(
                    [
                        (bytes(short_from_long_id(long_id)), long_id)
                        for long_id in get_latest_blocks(
                            self.blockchain_id, long_ids=True
                        )
                    ]
                )
                block_ids = [
                    block_long_ids[bytes(short_id)] for short_id in block_ids
                ]

            self._blocks = BlocksList.from_block_ids(
                block_ids, BlockLazilyLoaded
            )
        self._blocklist_lock.release()

    @property
    def block_ids(self) -> list[bytearray]:
        """Get a list of this blockchain's blocks' short IDs.

        Retained for backwards-compatibilty.
        """
        print(
            "`Blockchain.block_ids` has been deprecated because it returns "
            "block's short IDs. Use `Blockchain.get_block_ids()` instead, "
            "which returns their long IDs."
        )
        return self._blocks.get_short_ids()

    def terminate(self) -> None:
        """Clean up all resources this object uses."""
        self._terminate = True
        logger.debug("Terminating...")
        if self._blocks_listener:
            self._blocks_listener.terminate()
        logger.debug("Terminated!")

    def __del__(self) -> None:
        """Clean up all resources this object uses."""
        self.terminate()

    def delete(self) -> None:
        """Stop our participation in this blockchain.

        Locally deletes all data we have stored for it.
        """
        self.terminate()
        delete_blockchain(self.blockchain_id)

    @staticmethod
    def create(
        blockchain_name: str = "",
        block_received_handler: Callable[[Block], None] | None = None,
        app_name: str = "",
        appdata_dir: str = "",
        auto_load_missed_blocks: bool = True,
        forget_appdata: bool = False,
        sequential_block_handling: bool = True,
        update_blockids_before_handling: bool = False,
    ) -> "Blockchain":
        """Create and run a new blockchain.

        Args:
            blockchain_name (str):
                blockchain's local name: a human-readable label
                to ease recognition when manually interacting with WalytisAPI.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.
            block_received_handler (Callable):
                A function to be called every
                time a new block is received on this blockchain.
                This function should not have a long execution time because
                this class waits for its execution to finish before saving
                the received block in its list of processed blocks.
                If this handler raises an exception, it will be called
                with the same block at a later stage.
                The same block is never passed to this handler twice, and
                no child block will ever be passed to this function before
                its parent block.
            app_name (str): the unique name of this application, to distinguish
                between different applications that use the same blockchain
                for appdata management
            appdata_dir (str): the directory in which to save data on which
                    blocks this application has processed for this blockchain
            auto_load_missed_blocks (bool): whether or not to automatically get
                and process the blocks received by the blockchain while
                this application was offline.
            forget_appdata (bool): whether or not to ignore and overwrite any
                existing records of which blocks applications with the
                providede `app_name` have processed
            sequential_block_handling (bool): if True, the provided block-
                received handler will be executed on the main thread, and
                the handler will only be executed for the next block
                after the handler for the last block executes witout
                error.
                If set to False, the block-received handler will always be
                started on a new thread, meaning it is possible for
                multiple calls of the handler to be running in parallel
                for different blocks.
            update_blockids_before_handling (bool): whether or not the
                `blocks` attribute should be updated before running
                `block_received_handler` when a new block is received

        Returns:
            Blockchain: a Blockchain object representing the created blockchain
        """
        blockchain_id = create_blockchain(blockchain_name=blockchain_name)
        return Blockchain(
            blockchain_id,
            block_received_handler=block_received_handler,
            app_name=app_name,
            appdata_dir=appdata_dir,
            auto_load_missed_blocks=auto_load_missed_blocks,
            forget_appdata=forget_appdata,
            sequential_block_handling=sequential_block_handling,
            update_blockids_before_handling=update_blockids_before_handling,
        )

    @staticmethod
    def join(
        invitation: str,
        blockchain_name: str = "",
        block_received_handler: Callable[[Block], None] | None = None,
        app_name: str = "",
        appdata_dir: str = "",
        auto_load_missed_blocks: bool = True,
        forget_appdata: bool = False,
        sequential_block_handling: bool = True,
        update_blockids_before_handling: bool = False,
    ) -> "Blockchain":
        """Join a blockchain using an invitation generated by another node.

        Args:
            invitation (str): the invitation generated by a blockchain member
            blockchain_name (str):
                blockchain's local name: a human-readable label
                to ease recognition when manually interacting with WalytisAPI.
                Note: A blockchain's name only exists in the scope of a single
                node, meaning it is not guaranteed to be globally unique.
                Applications should use a blockchain's ID (attribute 'id') as
                its identifier, not the blockchain_name.
            block_received_handler (Callable):
                A function to be called every
                time a new block is received on this blockchain.
                This function should not have a long execution time because
                this class waits for its execution to finish before saving
                the received block in its list of processed blocks.
                If this handler raises an exception, it will be called
                with the same block at a later stage.
                The same block is never passed to this handler twice, and
                no child block will ever be passed to this function before
                its parent block.
            app_name (str): the unique name of this application, to distinguish
                between different applications that use the same blockchain
                for appdata management
            appdata_dir (str): the directory in which to save data on which
                    blocks this application has processed for this blockchain
            auto_load_missed_blocks (bool): whether or not to automatically get
                and process the blocks received by the blockchain while
                this application was offline.
            forget_appdata (bool): whether or not to ignore and overwrite any
                existing records of which blocks applications with the
                providede `app_name` have processed
            sequential_block_handling (bool): if True, the provided block-
                received handler will be executed on the main thread, and
                the handler will only be executed for the next block
                after the handler for the last block executes witout
                error.
                If set to False, the block-received handler will always be
                started on a new thread, meaning it is possible for
                multiple calls of the handler to be running in parallel
                for different blocks.
            update_blockids_before_handling (bool): whether or not the
                `blocks` attribute should be updated before running
                `block_received_handler` when a new block is received

        Returns:
            Blockchain: a Blockchain object representing the joined blockchain
        """
        blockchain_id = join_blockchain(
            invitation, blockchain_name=blockchain_name
        )
        return Blockchain(
            blockchain_id,
            block_received_handler=block_received_handler,
            app_name=app_name,
            appdata_dir=appdata_dir,
            auto_load_missed_blocks=auto_load_missed_blocks,
            forget_appdata=forget_appdata,
            sequential_block_handling=sequential_block_handling,
            update_blockids_before_handling=update_blockids_before_handling,
        )

    @staticmethod
    def from_blockchain_id(
        blockchain_id: str,
        block_received_handler: Callable[[Block], None] | None = None,
        app_name: str = "",
        appdata_dir: str = "",
        auto_load_missed_blocks: bool = True,
        forget_appdata: bool = False,
        sequential_block_handling: bool = True,
        update_blockids_before_handling: bool = False,
    ) -> "Blockchain":
        """Create an object to represent a Walytis blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain
            block_received_handler (Callable): function to be called every
                time a new block is received on this blockchain.
                This function should not have a long execution time because
                this class waits for its execution to finish before saving
                the received block in its list of processed blocks.
                If this eventhandler raises an exception, it will be called
                with the same block at a later stage.
                The same block is never passed to this handler twice, and
                no child block will ever be passed to this function before
                its parent block.
            app_name (str): the unique name of this application, to distinguish
                between different applications that use the same blockchain
                for appdata management
            appdata_dir (str): the directory in which to save data on which
                blocks this application has processed for this blockchain
            auto_load_missed_blocks (bool): whether or not to automatically get
                and process the blocks received by the blockchain while
                this application was offline.
            forget_appdata (bool): whether or not to ignore and overwrite any
                existing records of which blocks applications with the
                providede `app_name` have processed
            sequential_block_handling (bool): if True, the provided block-
                received handler will be executed on the main thread, and
                the handler will only be executed for the next block
                after the handler for the last block executes witout
                error.
                If set to False, the block-received handler will always be
                started on a new thread, meaning it is possible for
                multiple calls of the handler to be running in parallel
                for different blocks.
            update_blockids_before_handling (bool): whether or not the
                `blocks` attribute should be updated before running
                `block_received_handler` when a new block is received
        """
        return Blockchain(
            blockchain_id=blockchain_id,
            block_received_handler=block_received_handler,
            app_name=app_name,
            appdata_dir=appdata_dir,
            auto_load_missed_blocks=auto_load_missed_blocks,
            forget_appdata=forget_appdata,
            sequential_block_handling=sequential_block_handling,
            update_blockids_before_handling=update_blockids_before_handling,
        )



"""Testing functions to check GenericBlockchain implementations' integrity."""
import string
import random
from random import randint
from walytis_beta_api._experimental.generic_blockchain import GenericBlockchain, GenericBlock
from walytis_beta_api import Blockchain

# set the range of topic-lengths to test when adding blocks,
# from 0 to this value
N_BLOCK_TOPICS_TO_TEST = 4




def _generate_random_bytes():
    data_length = randint(1, 100)
    data = bytearray([0]) * data_length
    for i in range(data_length):
        data[i] = randint(0, 255)
    return data


def _generate_random_string(length=8) -> str:
    # Choose from lowercase and uppercase letters
    # letters = string.ascii_letters  # This is a-zA-Z
    letters = string.printable  # all printable characters
    return ''.join(random.choice(letters) for _ in range(length))


def _test_add_block(blockchain: GenericBlockchain, n_topics: int) -> GenericBlock:
    content = _generate_random_bytes()

    topics = [_generate_random_string(randint(1, 5)) for i in range(n_topics)]
    block_1 = blockchain.add_block(content, topics)

    assert blockchain.get_block_ids()[-1] == block_1.long_id, f"NT: {n_topics} - Blockchain.get_block_ids"
    assert blockchain.get_block(-1).short_id == block_1.short_id, f"NT: {n_topics} - Blockchain.get_block"
    assert blockchain.get_block(-1).content == block_1.content, f"NT: {n_topics} - Block.content"
    assert blockchain.get_block(-1).topics == block_1.topics, f"NT: {n_topics} - Block.topics"
    return block_1


def run_generic_blockchain_test(blockchain_type, **kwargs) -> GenericBlockchain:
    if not issubclass(blockchain_type, GenericBlockchain):
        raise ValueError(
            "The parameter `blockchain_type` must be a class that inherits "
            "`walytis_beta_api._experimental.generic_blockchain.GenericBlockchain`"
        )
    blockchain: GenericBlockchain = blockchain_type(**kwargs)
    # assert blockchain.get_num_blocks() == 0, "Provided blockchain shouldn't expose blocks yet"
    blocks = [
        _test_add_block(blockchain, i) for i in range(N_BLOCK_TOPICS_TO_TEST)
    ]

    # check that the last few block IDs are the created blocks
    success = True
    for i, block in enumerate(blocks):
        expected_index = i - len(blocks)
        if blockchain.get_block_ids()[expected_index] != block.long_id:
            success = False

    assert success, "No hidden blocks exposed"
    long_block_ids = blockchain.get_block_ids()
    blockchain.terminate()
    blockchain: GenericBlockchain = blockchain_type(**kwargs)
    assert long_block_ids == blockchain.get_block_ids(), "block_ids correctly reloaded"

    return blockchain




"""The library for interacting with Walytis.

This package contains the blockchain-application developer's API for
interacting with Walytis, as well as utilities common to both Walytis and
`walytis_beta_api`.
"""

from .walytis_beta_interface import (
    list_blockchains,
    list_blockchain_names,
    list_blockchain_ids,
    get_blockchain_id,
    get_blockchain_name,
    create_blockchain,
    delete_blockchain,
    join_blockchain,
    join_blockchain_from_zip,
    join_blockchain_from_cid,
    create_block,
    get_block,
    is_block_known,
    get_latest_blocks,
    get_blockchain_data,
    create_invitation,
    get_invitations,
    delete_invitation,
    get_walytis_beta_version,
    get_walytis_beta_version_string,
    get_walytis_beta_api_version,
    get_walytis_beta_api_version_string,
    BlocksListener,
    get_and_read_block,
    read_block,
)
from walytis_beta_tools.log import logger
from brenthy_tools_beta.brenthy_api import (
    get_brenthy_version,
    get_brenthy_version_string,
)
from .blockchain_model import Blockchain
from walytis_beta_tools.block_model import (
    Block,
    decode_short_id,
    decode_long_id,
    short_from_long_id,
)
from walytis_beta_tools.exceptions import (
    WalytisReplyDecodeError,
    WalytisBugError,
    NoSuchBlockchainError,
    BlockchainAlreadyExistsError,
    NoSuchInvitationError,
    JoinFailureError,
    BlockNotFoundError,
    BlockCreationError,
    BlockIntegrityError,
)
from walytis_beta_tools.versions import (
    WALYTIS_BETA_PROTOCOL_VERSION,
    WALYTIS_BETA_API_PROTOCOL_VERSION,
    WALYTIS_BETA_API_VERSION,
    WALYTIS_BETA_CORE_VERSION,
)



"""The versions of the the various Walytis components.

See `walytis_beta_tools.versions`
"""
from walytis_beta_tools.versions import (
    WALYTIS_BETA_PROTOCOL_VERSION,
    WALYTIS_BETA_API_PROTOCOL_VERSION,
    WALYTIS_BETA_API_VERSION,
    WALYTIS_BETA_CORE_VERSION,
)




"""Interfaces for `Block` and `Blockchain` classes for blockchain overlays."""

from abc import ABC, abstractmethod, abstractproperty
from typing import Callable, Iterable

from walytis_beta_tools._experimental.generic_block import GenericBlock


class GenericBlockchain(ABC):
    @abstractproperty
    def blockchain_id(self) -> str:
        pass

    @abstractproperty
    def block_received_handler(self) -> Callable[[GenericBlock], None] | None:
        pass

    @abstractmethod
    def add_block(
        self, content: bytes, topics: list[str] | str | None = None
    ) -> GenericBlock:
        pass

    @abstractmethod
    def get_blocks(self, reverse: bool = False) -> Iterable[GenericBlock]:
        pass

    @abstractmethod
    def get_block_ids(self) -> Iterable[bytes]:
        pass

    @abstractmethod
    def get_num_blocks(self) -> int:
        pass

    @abstractmethod
    def get_block(self, id: bytes) -> GenericBlock:
        pass

    @abstractmethod
    def get_peers(self) -> list[str]:
        pass

    @abstractmethod
    def terminate(self) -> None:
        pass

    @abstractmethod
    def delete(self) -> None:
        pass


class _GenericBlockchainImpl(GenericBlockchain):
    """A GenericBlockchain class with all abstract properties already declared.

    Inherit from this to reduce boilerplate code in your own child class, but
    you become responsible for remembering to assign values to all the fields!
    """

    def __init__(self):
        self.blockchain_id = ""

        self._block_received_handler = None

    @property
    def blockchain_id(self) -> str:
        return self._blockchain_id

    @blockchain_id.setter
    def blockchain_id(self, blockchain_id: str) -> None:
        self._blockchain_id = blockchain_id

    @property
    def block_received_handler(self) -> Callable[[GenericBlock], None] | None:
        return self._block_received_handler

    @block_received_handler.setter
    def block_received_handler(
        self, block_received_handler: Callable[[GenericBlock], None] | None
    ) -> None:
        self._block_received_handler = block_received_handler

    @property
    def block_ids(self) -> list[bytearray]:
        return self._block_ids



"""Interfaces for `Block` and `Blockchain` classes for blockchain overlays."""

from abc import ABC, abstractclassmethod, abstractproperty
from datetime import datetime
from datetime import timezone


class GenericBlock(ABC):
    @abstractclassmethod
    def from_id(
        cls, long_id: bytearray
    ):
        pass

    @abstractproperty
    def ipfs_cid(self) -> str:
        pass

    @abstractproperty
    def short_id(self) -> bytearray:
        pass

    @abstractproperty
    def long_id(self) -> bytearray:
        pass

    @abstractproperty
    def creator_id(self) -> bytearray:
        pass

    @abstractproperty
    def creation_time(self) -> datetime:
        pass

    @abstractproperty
    def topics(self) -> list[str]:
        pass

    @abstractproperty
    def content(self) -> bytearray:
        pass

    @abstractproperty
    def parents(self) -> list[bytearray]:
        pass

    @abstractproperty
    def file_data(self) -> bytearray:
        pass


class _GenericBlockImpl(GenericBlock):
    """A GenericBlock class with all abstract properties already declared.

    Inherit from this to reduce boilerplate code in your own child class, but
    you become responsible for remembering to assign values to all the fields!
    """

    def __init__(self):
        """Initialise an empty block object, not yet valid on a blockchain."""
        self._ipfs_cid = ""
        self._short_id = bytearray()
        self._long_id = bytearray()
        self._creator_id = bytearray()
        self._creation_time = datetime.now(timezone.utc)
        self._topics = []
        self._content = bytearray()
        self._parents = []
        self._file_data = bytearray()

    @property
    def ipfs_cid(self) -> str:
        return self._ipfs_cid

    @property
    def short_id(self) -> bytearray:
        return self._short_id

    @property
    def long_id(self) -> bytearray:
        return self._long_id

    @property
    def creator_id(self) -> bytearray:
        return self._creator_id

    @property
    def creation_time(self) -> datetime:
        return self._creation_time

    @property
    def topics(self) -> list[str]:
        return self._topics

    @property
    def content(self) -> bytearray:
        return self._content

    @property
    def parents(self) -> list[bytearray]:
        return self._parents

    @property
    def file_data(self) -> bytearray:
        return self._file_data



import logging
import os
from logging.handlers import RotatingFileHandler

# Formatter
# formatter = logging.Formatter(
#     '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
# )
MAX_RECORD_NAME_LENGTH = 16


class AlignedFormatter(logging.Formatter):
    def format(self, record):
        # Save original levelname for internal use
        original_levelname = record.levelname
        if len(record.name) > MAX_RECORD_NAME_LENGTH:
            record.name = record.name[:MAX_RECORD_NAME_LENGTH]
        # Format: [LEVEL] + padding (outside brackets)
        padded_level = f"[{original_levelname}]" + " " * (
            10 - len(original_levelname)
        )
        record.padded_level = padded_level

        return super().format(record)


formatter = AlignedFormatter(
    f"%(asctime)s %(padded_level)s %(name)-{
        MAX_RECORD_NAME_LENGTH
    }s: %(message)s"
)
# Console handler (INFO+)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(formatter)

LOG_PATH = os.path.abspath(
    os.environ.get("WALYTIS_BETA_LOG_PATH", "Walytis_Beta.log")
)
if os.path.isdir(LOG_PATH):
    LOG_PATH = os.path.join(LOG_PATH, "Walytis_Beta.log")
if not os.path.exists(os.path.dirname(LOG_PATH)):
    os.makedirs(LOG_PATH)
print(f"Walytis_Beta: Logging to {LOG_PATH}")
# File handler (DEBUG+ with rotation)
file_handler = RotatingFileHandler(
    LOG_PATH, maxBytes=5 * 1024 * 1024, backupCount=5
)
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(formatter)


# Log level IMPORTANT
IMPORTANT_LEVEL_NUM = 25
logging.addLevelName(IMPORTANT_LEVEL_NUM, "IMPORTANT")


def important(self, message, *args, **kwargs):
    if self.isEnabledFor(IMPORTANT_LEVEL_NUM):
        self._log(IMPORTANT_LEVEL_NUM, message, args, **kwargs)


logging.Logger.important = important


# # Root logger
# logger_root = logging.getLogger()
# logger_root.setLevel(logging.DEBUG)  # Global default
# logger_root.addHandler(console_handler)
# # logger_root.addHandler(file_handler)

logger = logging.getLogger("Walytis")
logger.setLevel(logging.DEBUG)


logger_networking = logging.getLogger("Walytis.Networking")
logger_api = logging.getLogger("Walytis.API")
logger_block_creation = logging.getLogger("Walytis.BlkCreation")
logger_block_processing = logging.getLogger("Walytis.BlkProcess")
logger_block_records = logging.getLogger("Walytis.BlkRecords")
logger_blockchain_model = logging.getLogger("Walytis.BcnModel")
logger_block_model = logging.getLogger("Walytis.BlkModel")
logger_generics = logging.getLogger("Walytis.Generics")
logger_ancestry = logging.getLogger("Walytis.Ancestry")
logger_appdata = logging.getLogger("Walytis.Appdata")
logger_join = logging.getLogger("Walytis.Join")

logger.setLevel(IMPORTANT_LEVEL_NUM)
logger_networking.setLevel(IMPORTANT_LEVEL_NUM)
logger_api.setLevel(IMPORTANT_LEVEL_NUM)
logger_block_creation.setLevel(IMPORTANT_LEVEL_NUM)
logger_block_processing.setLevel(IMPORTANT_LEVEL_NUM)
logger_block_records.setLevel(IMPORTANT_LEVEL_NUM)
logger_blockchain_model.setLevel(IMPORTANT_LEVEL_NUM)
logger_block_model.setLevel(IMPORTANT_LEVEL_NUM)
logger_generics.setLevel(IMPORTANT_LEVEL_NUM)
logger_ancestry.setLevel(IMPORTANT_LEVEL_NUM)
logger_appdata.setLevel(IMPORTANT_LEVEL_NUM)
logger_join.setLevel(IMPORTANT_LEVEL_NUM)

logger.addHandler(console_handler)
logger_networking.addHandler(console_handler)
logger_api.addHandler(console_handler)
logger_block_creation.addHandler(console_handler)
logger_block_processing.addHandler(console_handler)
logger_block_records.addHandler(console_handler)
logger_blockchain_model.addHandler(console_handler)
logger_block_model.addHandler(console_handler)
logger_generics.addHandler(console_handler)
logger_ancestry.addHandler(console_handler)
logger_appdata.addHandler(console_handler)
logger_join.addHandler(console_handler)

logger.addHandler(file_handler)
logger_networking.addHandler(file_handler)
logger_api.addHandler(file_handler)
logger_block_creation.addHandler(file_handler)
logger_block_processing.addHandler(file_handler)
logger_block_records.addHandler(file_handler)
logger_blockchain_model.addHandler(file_handler)
logger_block_model.addHandler(file_handler)
logger_generics.addHandler(file_handler)
logger_ancestry.addHandler(file_handler)
logger_appdata.addHandler(file_handler)
logger_join.addHandler(file_handler)



"""The machinery for communicating with Brenthy Core.

It contains functions that are a direct part of Walytis_Beta API and intended
for use by the Walytis_Beta API user (e.g. the create_blockchain() function),
but also functions that are used only by higher-level WalytisAPI constructs
such as BlocksListener class, which is not intended for use by the Walytis API
user, but is used by the blockchain_model.Blockchain class instead,
which is in turn intended for use by the Walytis_Beta API user.
"""

from __future__ import annotations
from walytis_beta_tools.log import logger_api as logger

import json
from datetime import datetime
from typing import Callable, Type

from brenthy_tools_beta import brenthy_api
from brenthy_tools_beta.utils import (
    bytes_to_string,
    decode_timestamp,
    from_b255_no_0s,
    function_name,
    string_to_bytes,
    to_b255_no_0s,
)
from brenthy_tools_beta.version_utils import (
    decode_version,
    encode_version,
)
from walytis_beta_tools._experimental.ipfs_interface import ipfs
from walytis_beta_tools.block_model import (
    Block,
    decode_short_id,  # pylint: disable=unused-import
    short_from_long_id,
)
from walytis_beta_tools.exceptions import (
    BlockchainAlreadyExistsError,
    BlockIntegrityError,
    BlockNotFoundError,
    JoinFailureError,
    NoSuchBlockchainError,
    NoSuchInvitationError,
    WalytisBugError,
    WalytisReplyDecodeError,
)
from walytis_beta_tools.versions import (
    WALYTIS_BETA_API_PROTOCOL_VERSION,
    WALYTIS_BETA_API_VERSION,
)

from .walytis_beta_generic_interface import (
    BaseBlocksListener,
    BaseWalytisBetaInterface,
    classproperty,
)

# storing this blockchain's name so that we don't missspell it
WALYTIS_BETA = "Walytis_Beta"


# --------------Settings ---------------------------
# logger.PRINT_DEBUG = False

NO_SUCH_BLOCKCHAIN_MESSAGE = "no such blockchain"
NO_SUCH_INVITATION_MESSAGE = "no such join-key"
BLOCK_NOT_FOUND = "block not found"
WALYTIS_BETA_ERROR_MESSAGE = "internal Walytis_Beta error"
BLOCKCHAIN_EXISTS_MESSAGE = "blockchain already exists"

# -------------- User Functions --------------------

# --- Blockchain IDs and Names ---


class _NetBlocksListener(BaseBlocksListener):
    """Listen for a blockchain's new block events."""

    event_listener = None

    def __init__(
        self,
        blockchain_id: str,
        eventhandler: Callable[[Block], None],
        topics: list[str] | None = None,
    ):
        """Call an eventhandler when the specified blockchain gets a new block.

        Args:
            blockchain_id (str): the id or name of the blockchain to listen on
            eventhandler (Callable[[Block, str], None]): eventhandler to call
                    when the blockchain gets a new block
            topics (list): the name of the topic to receive blocks from
        """
        blockchain_id = WalytisBetaNetApi.get_blockchain_id(blockchain_id)
        logger.info(f" BlocksListener: {blockchain_id}")
        self._blockchain_id = blockchain_id
        self._eventhandler = eventhandler
        if isinstance(topics, str):
            topics = [topics]
        if topics:
            self._topics = set(topics)
        else:
            self._topics = set()

        self.event_listener = brenthy_api.EventListener(
            "Walytis_Beta", self._bap_event_received,
            f"{blockchain_id}-NewBlocks"
        )

    def _bap_event_received(self, data: dict, topic: str) -> None:
        self._on_event_received(data, set([topic]))

    def get_block(self, block_id: bytearray) -> Block:
        return WalytisBetaNetApi.get_block(self.blockchain_id, block_id)

    @property
    def topics(self) -> set[str]:
        return self._topics

    @property
    def blockchain_id(self) -> str:
        return self._blockchain_id

    @property
    def eventhandler(self) -> Callable[[Block], None]:
        return self._eventhandler

    def terminate(self) -> None:
        """Clean up all resources used by this object."""
        if self.event_listener:
            self.event_listener.terminate()

    def __del__(self) -> None:
        """Clean up all resources used by this object."""
        self.terminate()


class WalytisBetaNetApi(BaseWalytisBetaInterface):
    @classproperty
    def BlocksListener(cls) -> Type[_NetBlocksListener]:
        return _NetBlocksListener

    @classmethod
    def list_blockchains(cls, names_first: bool = False) -> list[tuple[str, str]]:
        """Get the IDs names of all the blockchains we have.

        Args:
            names_first: determines the order of the the tuple contents in the
                            return value. See Returns.

        Returns:
            list: A list of tuples of the names and IDs of all the Walytis_Beta
                    blockchains we have. Tuples are (id, name) by default,
                    but (name, id) if names_first == True
        """
        # make request to Brenthy
        _response = cls._send_request("list_blockchains", bytearray([]))

        # read and process response from Brenthy
        try:
            response = json.loads(_response.decode())

            # handle unexpected response type
            if not isinstance(response, list):
                _error = cls._read_appcomterm_error_message(response)
                logger.error(f"WAPI: {function_name()}: {str(_error)}")
                raise _error

            if names_first:
                return [(name, id) for id, name in response]
            else:
                # convert lists to tuples
                return [(id, name) for id, name in response]

        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    # --- Blockchain Lifecycle ---

    @classmethod
    def create_blockchain(cls, blockchain_name: str = "") -> str:
        """Create and run a new blockchain.

        Args:
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with Walytis_API.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.

        Returns:
            str: ID of the created blockchain
        """
        payload = json.dumps({"blockchain_name": blockchain_name}).encode()

        _response = cls._send_request("create_blockchain", payload)
        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            success = response["success"]
        except Exception:
            error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if not success:
            error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

        blockchain_id = response["blockchain_id"]

        return blockchain_id

    @classmethod
    def join_blockchain(cls, invitation: str, blockchain_name: str = "") -> str:
        """Join a blockchain using an invitation generated by a blockchain node.

        Args:
            invitation (str): the invitation generated by a blockchain member
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with WalytisAPI.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.

        Returns:
            str: the ID of the joined blockchain
        """
        # convert invitation from JSON string to dict
        if isinstance(invitation, str):
            invitation = json.loads(invitation)
        if not isinstance(invitation, dict):
            raise ValueError(
                "The parameter invitation must be of type str or dict,"
                + f" not {type(invitation)}"
            )
        # logger.debug(f"WAPI: {function_name()}: ")
        # invitation is now of type dict
        blockchain_id = invitation["blockchain_id"]

        blockchains = cls.list_blockchains()

        if blockchain_id in [id for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(blockchain_id=blockchain_id)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if blockchain_name in [name for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(
                blockchain_name=blockchain_name)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        logger.debug(f"WAPI: {function_name()}: Making request...")

        response = json.loads(
            cls._send_request(
                "join_blockchain",
                json.dumps(
                    {
                        "invitation": invitation,
                        "blockchain_name": blockchain_name,
                    }
                ).encode(),
            ).decode()
        )
        logger.debug(f"WAPI: {function_name()}: Got response!")

        if response["success"]:
            return blockchain_id
        else:
            error = ""
            if "error" in response.keys():
                error = response["error"]
            error = JoinFailureError(error_message=error)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

    @classmethod
    def join_blockchain_from_cid(cls,
                                 blockchain_id: str, blockchain_data_cid: str, blockchain_name: str = ""
                                 ) -> None:
        """Join an existing live blockchain, given the IPFS CID of its appdata.

        Args:
            blockchain_id (str): the name of the blockchain
            blockchain_data_cid (str): IPFS CID of the zip file containing the
                    blockchain's data
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with WalytisAPI.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.
        """
        blockchains = cls.list_blockchains()

        if blockchain_id in [id for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(blockchain_id=blockchain_id)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if blockchain_name in [name for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(
                blockchain_name=blockchain_name)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

        request = json.dumps(
            {
                "blockchain_id": blockchain_id,
                "blockchain_name": blockchain_name,
                "blockchain_data_cid": blockchain_data_cid,
            }
        ).encode()
        logger.debug(f"WAPI: {function_name()}: Making request...")

        _response = cls._send_request("join_blockchain_from_cid", request)
        logger.debug(f"WAPI: {function_name()}: Got Response!")

        try:
            response = json.loads(_response.decode())
            success = response["success"]
        except Exception:
            _error = cls._read_appcomterm_error_message(
                response, default_error=JoinFailureError()
            )
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error
        if not success:
            raise JoinFailureError

    @classmethod
    def delete_blockchain(cls, blockchain_id: str) -> None:
        """Delete the blockchain with the given id or name.

        Args:
            blockchain_id (str): the id or name of the blockchain to delete
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        payload = blockchain_id.encode()

        _response = cls._send_request("delete_blockchain", payload)

        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            success = response["success"]
        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error
        if not success:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def create_block(cls,
                     blockchain_id: str,
                     content: bytearray | bytes,
                     topics: list[str] | str | None = None
                     ) -> Block:
        """Create a block with the given content on the specified blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to create a block
            content (bytearray): the content to store on the blockchain
            topics (list): user's metadata to encode in the block's ID
        Returns:
            Block: the newly created block
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        if not (isinstance(content, (bytearray, bytes))):
            error = ValueError(
                "Block content must be of type bytes or bytearray.")
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

        if isinstance(topics, str):
            topics = [topics]
        if topics is None:
            topics = []

        response = {}

        data = blockchain_id.encode() + bytearray([0])

        data += to_b255_no_0s(len(topics)) + bytearray([0])
        for topic in topics:
            data = data + topic.encode() + bytearray([0, 0, 0])

        # adding the self's content
        data = data + content

        # askig the blockchain to build and publish the block
        response = json.loads(cls._send_request("create_block", data).decode())

        # if the blockchain says it failed to build the block
        if not response.get("success"):
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

        block = cls.get_and_read_block(
            string_to_bytes(response["block_id"])
        )  # decode the block

        return block

    @classmethod
    def get_block(cls, blockchain_id: str, block_id: bytearray) -> Block:
        """Get a block from the specified blokchain given its block ID.

        Args:
            blockchain_id (str): the id or name of the blockchain from which to get
                                    the block
            block_id (str): the id of the block
        Returns:
            Block: the Block object for the requested block
        """
        if not blockchain_id or not block_id:
            error = ValueError(
                f"The parameters can't be empty\n{blockchain_id}\n{block_id}"
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if not isinstance(blockchain_id, str):
            error_message = (
                " get_block: parameter blockchain_id must be of "
                f"type str, not {type(blockchain_id)}"
            )
            logger.error(error_message)
            raise TypeError(error_message)
        if not (isinstance(block_id, (bytearray, bytes))):
            error_message = (
                " get_block: parameter block_id must "
                f"be of type bytearray or bytes, not {type(block_id)}"
            )
            logger.error(error_message)
            raise TypeError(error_message)
        response = {}
        payload = json.dumps(
            {"blockchain_id": blockchain_id,
                "block_id": bytes_to_string(block_id)}
        ).encode()
        # askig the blockchain to build and publish the block
        response = json.loads(cls._send_request("get_block", payload).decode())

        # if the blockchain says it failed to build the block
        if not response.get("success"):
            _error = cls._read_appcomterm_error_message(
                response, default_error=BlockNotFoundError()
            )
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

        block = cls.get_and_read_block(
            string_to_bytes(response["block_id"])
        )  # decode the block

        return block

    @classmethod
    def is_block_known(cls, blockchain_id: str, block_id: bytearray) -> bool:
        """Check whether or not the specified blockchain has the given block.

        Args:
            blockchain_id (str): the id or name of the blockchain to which the
                                    block to look up belongs
            block_id (bytearray): the block ID to look up in the block ID records
        Returns:
            bool: Whether the block is a known block
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)

        # Sending request to Brenthy
        payload = blockchain_id.encode() + bytearray([0]) + block_id

        # askig the blockchain to build and publish the block
        _response = cls._send_request("is_block_known", payload)
        response = json.loads(_response.decode())

        # Reading response from Brenthy
        if not response.get("success"):
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error
        return response["is known"]

    @classmethod
    def get_latest_blocks(cls,
                          blockchain_id: str,
                          amount: int | None = None,
                          since_date: datetime | None = None,
                          topics: list[str] | None = None,
                          long_ids: bool = False,  # defaults to False for backwards-compatibility
                          ) -> list[bytearray]:
        """Get the latest few blocks on the blockchain.

        The amount of which is determined by the paramters.
        The blocks are returned in order of reception or creation date (CHECK!)
        from most recent to oldest.
        One of the paramaters amount or since_date must be set.

        Args:
            blockchain_id (str): the id or name of the blockchain to look up
            amount (int): how many blocks should be returned
            since_date (datetime): the maximum age of the blocks to look for.
                    This function will return several even older blocks,
                    some of which may have been received after the specified date
            topics (list): the topics to filter by, returning only
                    blocks published on those topics, if this paramater is set.
                    If not set, this function returns all blocks within the range
                    of the first two parameters

        Returns:
            list: a list of IDs of the latest few blocks
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)

        payload = json.dumps(
            {
                "blockchain_id": blockchain_id,
                "amount": amount,
                "since_date": since_date,
                "topics": topics,
                "long_ids": long_ids
            }
        ).encode()

        # send request to Brenthy
        _response = cls._send_request("get_latest_blocks", payload)

        # read response from Brenthy
        try:
            response = json.loads(_response.decode())
            block_ids_encoded = response.get("block_ids")
            block_ids = []
            for block_id in block_ids_encoded:
                block_ids.append(string_to_bytes(block_id))

            # BACKWARDS COMPATIBILITY foor versions of WalytisCore that don't
            # support the `long_ids` parameter:
            if bytearray([0, 0, 0, 0]) not in block_ids[0]:
                block_ids = [
                    cls._get_block_long_from_short_id(blockchain_id, short_id)
                    for short_id in block_ids
                ]

            return block_ids
        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def get_blockchain_data(cls, blockchain_id: str) -> str:
        """Create a zip file of all this blockchain's data, returning its path.

        Args:
            blockchain_id (str): the id or name of the blockchain to zip
        Returns:
            str: the path of the zip file created
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        request = blockchain_id.encode()

        _response = cls._send_request("get_blockchain_data", request)
        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            if response["success"]:
                return response["path"]

        except Exception:
            pass
        _error = cls._read_appcomterm_error_message(response)
        logger.error(f"WAPI: {function_name()}: {str(_error)}")
        raise _error

    @classmethod
    def get_peers(cls, blockchain_id: str) -> list[str]:
        """Get a list of IPFS IDs of the given blockchain's currently online nodes.

        Args:
            blockchain_id (str): the id or name of the blockchain to look up
        Returns:
            list[str]: the IPFS IDs of the blockchain's currently online nodes
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        request = blockchain_id.encode()

        _response = cls._send_request("get_peers", request)
        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            if response["success"]:
                return response["peers"]

        except Exception:
            pass
        _error = cls._read_appcomterm_error_message(response)
        logger.error(f"WAPI: {function_name()}: {str(_error)}")
        raise _error

    @classmethod
    def create_invitation(cls,
                          blockchain_id: str, one_time: bool = True, shared: bool = False
                          ) -> str:
        """Create a code which other nodes can use to join this blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to create a
                            invitation for
            one_time (bool): (optional, default True) whether this code is valid
                            only for one use only instead of being reusable
            shared (bool): whether this invitation should be hosted by other
                            nodes who join this blockchain too, or just by this one
        Returns:
            str: an invitation code that can be used by other computers to join
                    this blockchain
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        request = json.dumps(
            {
                "blockchain_id": blockchain_id,
                "one_time": one_time,
                "shared": shared,
            }
        ).encode()

        _response = cls._send_request("create_invitation", request)
        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            return response["invitation"]
        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def get_invitations(cls, blockchain_id: str) -> list[str]:
        """Get the specified blockchain's active invitations.

        Args:
            blockchain_id (str): the id or name of the blockchain whose invitations
                                    to look up
        Returns:
            list: the invitations of this blockchain
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)

        # Sending request to Brenthy
        payload = blockchain_id.encode()

        # askig the blockchain to build and publish the block
        _response = cls._send_request("get_invitations", payload)

        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            if response["success"]:
                return response["invitations"]
            else:
                _error = cls._read_appcomterm_error_message(response)
                logger.error(f"WAPI: {function_name()}: {str(_error)}")
                raise _error

        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def delete_invitation(cls, blockchain_id: str, invitation: str) -> None:
        """Delete an invitation from the specified blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to delete an
                                invitation from
            invitation (str): the invitation to delete
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        request = json.dumps(
            {"blockchain_id": blockchain_id, "invitation": invitation}
        ).encode()

        _response = cls._send_request("delete_invitation", request)
        # Reading response from Brenthy
        try:
            response = json.loads(_response.decode())
            if not response["success"]:
                _error = cls._read_appcomterm_error_message(response)
                logger.error(f"WAPI: {function_name()}: {str(_error)}")
                raise _error
        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def get_walytis_beta_version(cls,) -> tuple[int, int, int]:
        """Get the software version of the locally running Walytis node.

        Returns:
            tuple: the software version of the locally running Walytis_Beta node
        """
        # make request to Brenthy
        _response = cls._send_request(
            "get_walytis_beta_version", bytearray([]))

        # read and process response from Brenthy
        try:
            response = json.loads(_response.decode())

            # handle unexpected response type
            if not (
                isinstance(response, dict)
                and (response.get("walytis_beta_core_version"), list)
            ):
                _error = cls._read_appcomterm_error_message(response)
                logger.error(f"WAPI: {function_name()}: {str(_error)}")
                raise _error

            return tuple(response["walytis_beta_core_version"])

        except Exception:
            _error = cls._read_appcomterm_error_message(response)
            logger.error(f"WAPI: {function_name()}: {str(_error)}")
            raise _error

    @classmethod
    def get_walytis_beta_api_version(cls,) -> tuple[int, int, int]:
        """Get the software version of the this walytis_beta_api library.

        Returns:
            tuple: the software version of the this walytis_beta_api library
        """
        return WALYTIS_BETA_API_VERSION

    @classmethod
    def get_and_read_block(cls, short_id: bytearray) -> Block:
        """Given a block's ID, download and verify it.

        Args:
            short_id (str): the ID of the block to get
        Returns:
            Block: an object representing the specified block, containing all its
                    metadata and block content
        """
        try:
            ipfs_cid = decode_short_id(short_id)["ipfs_cid"]
            block_data = ipfs.files.read(ipfs_cid)
        except Exception:
            error = BlockNotFoundError()
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        try:
            block = cls.read_block(block_data, ipfs_cid)
        except Exception as error:
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        while short_id[-1] == 0:
            short_id = short_id[:-1]
        if bytearray(block.short_id) != short_from_long_id(short_id):  # suspicious
            error_message = (
                "The decoded block's ID is not the same as the one we are looking "
                f"for.\nLooking for: {bytearray(short_id)}"
                f"\nGenerated from BlockData: {block.short_id}"
            )

            exception = BlockIntegrityError(error_message)
            logger.error(f"WAPI: {function_name()}: {error_message}")
            raise exception
        return block

    @classmethod
    def read_block(cls, block_data: bytearray, ipfs_cid: str) -> Block:
        """Read a block-file, generating a block object if it's valid.

        Args:
            block_data (bytearray): the raw data of the block to read.
            ipfs_cid (str): the IPFS content ID of the block
        Returns:
            Block: a Block object
        """
        # making a copy of the data to work with
        data = bytearray(block_data)
        # extract metadata block (separated from content by [0,0,0,0,0])
        # and splt it into is subcomponents separated by [0,0]
        content_separator = data.index(bytearray([0, 0, 0, 0, 0]))
        blockfile_header = data[:content_separator].split(
            bytearray([0, 0, 0, 0]))
        metadata = blockfile_header[0].split(bytearray([0, 0]))
        # logger.important(data)
        # logger.important(blockfile_header)
        # logger.important(metadata)
        if len(blockfile_header) > 1:
            parents = blockfile_header[1].split(bytearray([0, 0, 0]))
        else:
            parents = []
            # logger.important("Genesis block")
        # content sits between metadata and block_hash
        content = data[content_separator + 5:]

        blockchain_version = decode_version(metadata[0])
        creator_id = metadata[1]
        creation_time = decode_timestamp(metadata[2])

        topics = [topic.decode()
                  for topic in metadata[3].split(bytearray([0]))]
        content_length = from_b255_no_0s(metadata[4])
        content_hash_algorithm = metadata[5].decode()
        content_hash = metadata[6]

        n_parents = from_b255_no_0s(metadata[7])
        parents_hash_algorithm = metadata[8].decode()
        parents_hash = metadata[9]

        # creating a new block object from the decoded block data
        block = Block.from_metadata(
            blockchain_version=blockchain_version,
            creator_id=creator_id,
            creation_time=creation_time,
            topics=topics,
            content_hash_algorithm=content_hash_algorithm,
            content_hash=content_hash,
            content=content,
            n_parents=n_parents,
            parents_hash_algorithm=parents_hash_algorithm,
            parents_hash=parents_hash,
            parents=parents,
            ipfs_cid=ipfs_cid,
            content_length=content_length,
            file_data=bytearray(block_data),
        )

        block.generate_id()

        # making sure the block's hash is correct
        if not block.check_integrity():
            error = BlockIntegrityError(
                "The received block had an incorrect hash. The block is "
                "considered corrupt or fraudulently manipulated."
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error

        return block

    @classmethod
    def _read_appcomterm_error_message(cls,
                                       data: dict | bytearray, default_error: BaseException | None = None
                                       ) -> BaseException:
        """Analyse a failed request reply, returning an appropriate Exception.

        Args:
            data (dict | bytearray): the response from the brenthy AppComTerm to
                                        analyse
            default_error(Exception): the exception to return if no error
                                        message/cause can be determined
        """
        if isinstance(data, dict):
            if "error" in data.keys():
                if data["error"] == NO_SUCH_BLOCKCHAIN_MESSAGE:
                    if "blockchain_id" in data.keys():
                        return NoSuchBlockchainError(
                            blockchain_id=data["blockchain_id"]
                        )
                    else:
                        return NoSuchBlockchainError()
                elif data["error"] == BLOCKCHAIN_EXISTS_MESSAGE:
                    if "blockchain_name" in data.keys():
                        return BlockchainAlreadyExistsError(
                            blockchain_name=data["blockchain_name"]
                        )
                    else:
                        return BlockchainAlreadyExistsError()
                elif data["error"] == NO_SUCH_INVITATION_MESSAGE:
                    return NoSuchInvitationError()
                elif data["error"] == WALYTIS_BETA_ERROR_MESSAGE:
                    return WalytisBugError(error_message=data["error"])
                if default_error:
                    return default_error
                else:
                    return WalytisReplyDecodeError(reply=str(data))
        if default_error:
            return default_error
        logger.error(str(WalytisReplyDecodeError(reply=str(data))))
        return WalytisReplyDecodeError(reply=str(data))

    @classmethod
    def _send_request(cls, function_name: str, payload: bytearray | bytes) -> bytearray:
        """Send a request to Walytis-Core.

        Args:
            function_name(str): the name of the function in walytis_api_terminal
                                    which we want to call
            payload(bytearray): the data the function in walytis_beta_api_terminal
                                    needs to process our request, its arguments

        Returns:
            bytearray: the reply from the function we called in
                                    walytis_beta_api_terminal
        """
        request = (
            encode_version(WALYTIS_BETA_API_PROTOCOL_VERSION) + bytearray([0])
            + function_name.encode() + bytearray([0])
            + payload
        )
        reply = brenthy_api.send_request("Walytis_Beta", request)
        walytis_api_version = decode_version(
            reply[: reply.index(bytearray([0]))])
        reply = reply[reply.index(bytearray([0])) + 1:]
        return reply

    @classmethod
    def _get_block_long_from_short_id(cls, blockchain_id: str, short_id: bytearray):
        """This function is needed for BACKWARDS COMPATIBILITY with older versions
        of Walytis_Beta Core.
        """
        return cls.get_block(blockchain_id, short_id).long_id



"""The fundamental machinery constituting Blocks.

See `walytis_beta_tools.block_model`
"""


import hashlib
from copy import deepcopy

# from random import seed
# from random import randint
from datetime import datetime
from secrets import randbits

from walytis_beta_tools.log import logger_block_model as logger
from brenthy_tools_beta.utils import (  # pylint: disable=unused-import
    bytes_from_b255_no_0s,
    bytes_to_b255_no_0s,
    decode_timestamp,
    encode_timestamp,
    from_b255_no_0s,
    to_b255_no_0s,
)
from brenthy_tools_beta.version_utils import decode_version, encode_version

from .exceptions import BlockIntegrityError, InvalidBlockIdError
from walytis_beta_tools._experimental.generic_block import _GenericBlockImpl as GenericBlock

PREFERRED_HASH_ALGORITHM = "sha512"

from walytis_beta_tools.block_model import (
    Block,
    decode_short_id,
    decode_long_id,
    short_from_long_id,
)
print("DEPRECATING: import from walytis_beta_api instead of walytis_beta_api.block_model")


"""The versions of the the various Walytis components."""

# the newest version of the Walytis_Beta protocol, which defines
# how nodes communicate and how block IDs and block-files are structured
WALYTIS_BETA_PROTOCOL_VERSION = 2

# the newest version of the communication protocol used by the Walytis node
# and the walytis_beta_api library to communicate with each other
WALYTIS_BETA_API_PROTOCOL_VERSION = (WALYTIS_BETA_PROTOCOL_VERSION, 4)

# the version of the walytis_api library which is used by applications
# to communicate with the Walytis_Beta node to interact with blockchains
WALYTIS_BETA_API_VERSION = (*WALYTIS_BETA_API_PROTOCOL_VERSION, 11)

# the version of the Walytis core software, which runs blockchains
WALYTIS_BETA_CORE_VERSION = (*WALYTIS_BETA_API_PROTOCOL_VERSION, 12)

WALYTIS_BETA_PACKAGE_VERSION = (
    *WALYTIS_BETA_API_PROTOCOL_VERSION,
    WALYTIS_BETA_API_VERSION[-1] + WALYTIS_BETA_CORE_VERSION[-1],
)

_WALYTIS_BETA_API_VERSION = ".".join(
    [str(v) for v in WALYTIS_BETA_API_VERSION]
)
_WALYTIS_BETA_PACKAGE_VERSION = ".".join(
    [str(v) for v in WALYTIS_BETA_PACKAGE_VERSION]
)



"""The machinery for communicating with Brenthy Core.

It contains functions that are a direct part of Walytis_Beta API and intended
for use by the Walytis_Beta API user (e.g. the create_blockchain() function),
but also functions that are used only by higher-level WalytisAPI constructs
such as BlocksListener class, which is not intended for use by the Walytis API
user, but is used by the blockchain_model.Blockchain class instead,
which is in turn intended for use by the Walytis_Beta API user.
"""

from __future__ import annotations

import json
import os
from datetime import datetime
from threading import Thread
from typing import Callable, Type, TypeVar

from brenthy_tools_beta.utils import (
    function_name,
    make_directory_readable,
    make_file_readable,
)
from walytis_beta import walytis_beta, walytis_beta_api_terminal
from walytis_beta_tools.block_model import (
    Block,
    decode_short_id,  # pylint: disable=unused-import
    short_from_long_id,
)
from walytis_beta_tools.exceptions import (
    BlockchainAlreadyExistsError,  # noqa
    BlockIntegrityError,  # noqa
    BlockNotFoundError,  # noqa
    JoinFailureError,  # noqa
    NoSuchBlockchainError,  # noqa
    NoSuchInvitationError,  # noqa
    WalytisBugError,  # noqa
    WalytisReplyDecodeError,  # noqa
)
from walytis_beta_tools.log import logger_api as logger
from walytis_beta_tools.versions import (
    WALYTIS_BETA_API_VERSION,
    WALYTIS_BETA_CORE_VERSION,
)

from .walytis_beta_generic_interface import (
    BaseBlocksListener,
    BaseWalytisBetaInterface,
    classproperty,
)


class _DirectBlocksListener(BaseBlocksListener):
    """Listen for a blockchain's new block events."""

    event_listener = None

    def __init__(
        self,
        blockchain_id: str,
        eventhandler: Callable[[Block], None],
        topics: list[str] | None = None,
    ):
        """Call an eventhandler when the specified blockchain gets a new block.

        Args:
            blockchain_id (str): the id or name of the blockchain to listen on
            eventhandler (Callable[[Block, str], None]): eventhandler to call
                    when the blockchain gets a new block
            topics (list): the name of the topic to receive blocks from
        """
        self._blockchain_id = WalytisBetaDirectInterface.get_blockchain_id(
            blockchain_id
        )
        self._eventhandler = eventhandler
        if isinstance(topics, str):
            topics = [topics]
        if topics:
            self._topics = set(topics)
        else:
            self._topics = set()
        walytis_beta_api_terminal.add_eventhandler(self._waly_event_received)

    def _waly_event_received(self, data: dict, topics: list[str]) -> None:
        Thread(
            target=self._on_event_received, args=(data, topics),
            name=f"_DirectBlocksListener.eventhandler-{self.blockchain_id}"
        ).start()

        # self._on_event_received(data,set(topics))
    @property
    def blockchain_id(self) -> str:
        return self._blockchain_id

    @property
    def topics(self) -> set[str]:
        return self._topics

    @property
    def eventhandler(self) -> Callable[[Block], None]:
        return self._eventhandler

    def get_block(self, block_id: bytearray) -> Block:
        return WalytisBetaDirectInterface.get_block(self.blockchain_id, block_id)

    def terminate(self) -> None:
        """Clean up all resources used by this object."""
        pass


TBlocksListener = TypeVar('TBlocksListener', bound=BaseBlocksListener)


class WalytisBetaDirectInterface(BaseWalytisBetaInterface):
    @ classproperty
    def BlocksListener(cls) -> Type[DirectBlocksListener]:
        return _DirectBlocksListener

    @ classmethod
    def list_blockchains(cls, names_first: bool = False) -> list[tuple[str, str]]:
        """Get the IDs names of all the blockchains we have.

        Args:
            names_first: determines the order of the the tuple contents in the
                            return value. See Returns.

        Returns:
            list: A list of tuples of the names and IDs of all the Walytis_Beta
                    blockchains we have. Tuples are (id, name) by default,
                    but (name, id) if names_first == True
        """
        if names_first:
            return [
                (blockchain.name, blockchain.blockchain_id)
                for blockchain in walytis_beta.blockchains

            ]
        else:
            return [
                (blockchain.blockchain_id, blockchain.name)
                for blockchain in walytis_beta.blockchains
            ]

    # --- Blockchain Lifecycle ---

    @ classmethod
    def create_blockchain(cls, blockchain_name: str = "") -> str:
        """Create and run a new blockchain.

        Args:
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with Walytis_API.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.

        Returns:
            str: ID of the created blockchain
        """
        blockchain = walytis_beta.create_blockchain(blockchain_name)
        if not blockchain:
            raise BlockchainAlreadyExistsError()
        return blockchain.blockchain_id

    @ classmethod
    def join_blockchain(cls, invitation: str, blockchain_name: str = "") -> str:
        """Join a blockchain using an invitation generated by a blockchain node.

        Args:
            invitation (str): the invitation generated by a blockchain member
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with WalytisAPI.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.

        Returns:
            str: the ID of the joined blockchain
        """
        if isinstance(invitation, str):
            invitation = json.loads(invitation)
        if not isinstance(invitation, dict):
            raise ValueError(
                "The parameter invitation must be of type str or dict,"
                + f" not {type(invitation)}"
            )
        # logger.debug(f"WAPI: {function_name()}: ")
        # invitation is now of type dict
        blockchain_id = invitation["blockchain_id"]
        blockchains = cls.list_blockchains()

        if blockchain_id in [id for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(blockchain_id=blockchain_id)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if blockchain_name in [name for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(
                blockchain_name=blockchain_name)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        logger.debug(f"WAPI: {function_name()}: Making request...")
        blockchain = walytis_beta.join_blockchain(
            invitation, blockchain_name=blockchain_name
        )
        if not blockchain:
            raise JoinFailureError()
        return blockchain.blockchain_id

    @ classmethod
    def join_blockchain_from_cid(cls,
                                 blockchain_id: str, blockchain_data_cid: str, blockchain_name: str = ""
                                 ) -> None:
        """Join an existing live blockchain, given the IPFS CID of its appdata.

        Args:
            blockchain_id (str): the name of the blockchain
            blockchain_data_cid (str): IPFS CID of the zip file containing the
                    blockchain's data
            blockchain_name (str): blockchain's local name: a human-readable label
                    to ease recognition when manually interacting with WalytisAPI.
                    Note: A blockchain's name only exists in the scope of a single
                    node, meaning it is not guaranteed to be globally unique.
                    Applications should use a blockchain's ID (attribute 'id') as
                    its identifier, not the blockchain_name.
        """
        blockchains = cls.list_blockchains()

        if blockchain_id in [id for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(blockchain_id=blockchain_id)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if blockchain_name in [name for id, name in blockchains]:
            error = BlockchainAlreadyExistsError(
                blockchain_name=blockchain_name)
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        blockchain = walytis_beta.join_blockchain_from_cid(
            blockchain_id, blockchain_data_cid, blockchain_name=blockchain_name
        )
        if not blockchain:
            raise JoinFailureError()
        return blockchain.blockchain_id

    @ classmethod
    def delete_blockchain(cls, blockchain_id: str) -> None:
        """Delete the blockchain with the given id or name.

        Args:
            blockchain_id (str): the id or name of the blockchain to delete
        """
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        walytis_beta.delete_blockchain(blockchain_id)

    @ classmethod
    def create_block(cls,
                     blockchain_id: str,
                     content: bytearray | bytes,
                     topics: list[str] | str | None = None
                     ) -> Block:
        """Create a block with the given content on the specified blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to create a block
            content (bytearray): the content to store on the blockchain
            topics (list): user's metadata to encode in the block's ID
        Returns:
            Block: the newly created block
        """
        if not (isinstance(content, (bytearray, bytes))):
            error = ValueError(
                "Block content must be of type bytes or bytearray.")
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        blockchain = cls._get_blockchain(blockchain_id)
        if topics is None:
            topics = []
        elif isinstance(topics, str):
            topics = [topics]
        block = blockchain.create_block(content, topics)
        return block

    @ classmethod
    def get_block(cls, blockchain_id: str, block_id: bytearray) -> Block:
        """Get a block from the specified blokchain given its block ID.

        Args:
            blockchain_id (str): the id or name of the blockchain from which to get
                                    the block
            block_id (str): the id of the block
        Returns:
            Block: the Block object for the requested block
        """
        if not blockchain_id or not block_id:
            error = ValueError(
                f"The parameters can't be empty\n{blockchain_id}\n{block_id}"
            )
            logger.error(f"WAPI: {function_name()}: {str(error)}")
            raise error
        if not isinstance(blockchain_id, str):
            error_message = (
                " get_block: parameter blockchain_id must be of "
                f"type str, not {type(blockchain_id)}"
            )
            logger.error(error_message)
            raise TypeError(error_message)
        if not (isinstance(block_id, (bytearray, bytes))):
            error_message = (
                " get_block: parameter block_id must "
                f"be of type bytearray or bytes, not {type(block_id)}"
            )
            logger.error(error_message)
            raise TypeError(error_message)

        blockchain = cls._get_blockchain(blockchain_id)
        if blockchain.is_block_known(block_id):
            ipfs_cid = decode_short_id(block_id)["ipfs_cid"]
            data_file = blockchain.get_block_datafile_path(block_id)
            with open(data_file, "rb") as file:
                block_data = file.read()
            block = blockchain.read_block(block_data, ipfs_cid, live=False)
        else:
            block = blockchain.download_and_process_block(block_id)
        if not block:
            raise BlockNotFoundError()
        return block

    @ classmethod
    def is_block_known(cls, blockchain_id: str, block_id: bytearray) -> bool:
        """Check whether or not the specified blockchain has the given block.

        Args:
            blockchain_id (str): the id or name of the blockchain to which the
                                    block to look up belongs
            block_id (bytearray): the block ID to look up in the block ID records
        Returns:
            bool: Whether the block is a known block
        """
        blockchain = cls._get_blockchain(blockchain_id)
        return blockchain.is_block_known(block_id)

    @ classmethod
    def get_latest_blocks(cls,
                          blockchain_id: str,
                          amount: int | None = None,
                          since_date: datetime | None = None,
                          topics: list[str] | None = None,
                          long_ids: bool = False,  # defaults to False for backwards-compatibility
                          ) -> list[bytearray]:
        """Get the latest few blocks on the blockchain.

        The amount of which is determined by the paramters.
        The blocks are returned in order of reception or creation date (CHECK!)
        from most recent to oldest.
        One of the paramaters amount or since_date must be set.

        Args:
            blockchain_id (str): the id or name of the blockchain to look up
            amount (int): how many blocks should be returned
            since_date (datetime): the maximum age of the blocks to look for.
                    This function will return several even older blocks,
                    some of which may have been received after the specified date
            topics (list): the topics to filter by, returning only
                    blocks published on those topics, if this paramater is set.
                    If not set, this function returns all blocks within the range
                    of the first two parameters

        Returns:
            list: a list of IDs of the latest few blocks
        """
        blockchain = cls._get_blockchain(blockchain_id)
        block_ids = blockchain.load_latest_block_ids(
            amount, since_date, topics
        )
        if long_ids:
            return block_ids
        else:
            return [
                short_from_long_id(block_id) for block_id in block_ids
            ]

    @ classmethod
    def get_blockchain_data(cls, blockchain_id: str) -> str:
        """Create a zip file of all this blockchain's data, returning its path.

        Args:
            blockchain_id (str): the id or name of the blockchain to zip
        Returns:
            str: the path of the zip file created
        """
        blockchain = cls._get_blockchain(blockchain_id)
        result = blockchain.zip_appdata()
        make_file_readable(result)
        make_directory_readable(os.path.dirname(result))
        return result

    @ classmethod
    def get_peers(cls, blockchain_id: str) -> list[str]:
        """Get a list of IPFS IDs of the given blockchain's currently online nodes.

        Args:
            blockchain_id (str): the id or name of the blockchain to look up
        Returns:
            list[str]: the IPFS IDs of the blockchain's currently online nodes
        """
        blockchain = cls._get_blockchain(blockchain_id)

        return blockchain.get_peers()

    @ classmethod
    def create_invitation(cls,
                          blockchain_id: str, one_time: bool = True, shared: bool = False
                          ) -> str:
        """Create a code which other nodes can use to join this blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to create a
                            invitation for
            one_time (bool): (optional, default True) whether this code is valid
                            only for one use only instead of being reusable
            shared (bool): whether this invitation should be hosted by other
                            nodes who join this blockchain too, or just by this one
        Returns:
            str: an invitation code that can be used by other computers to join
                    this blockchain
        """
        blockchain = cls._get_blockchain(blockchain_id)
        return blockchain.create_invitation(
            one_time=one_time, shared=shared
        )

    @ classmethod
    def get_invitations(cls, blockchain_id: str) -> list[str]:
        """Get the specified blockchain's active invitations.

        Args:
            blockchain_id (str): the id or name of the blockchain whose invitations
                                    to look up
        Returns:
            list: the invitations of this blockchain
        """
        blockchain = cls._get_blockchain(blockchain_id)
        return blockchain.invitations

    @ classmethod
    def delete_invitation(cls, blockchain_id: str, invitation: str) -> None:
        """Delete an invitation from the specified blockchain.

        Args:
            blockchain_id (str): the id or name of the blockchain to delete an
                                invitation from
            invitation (str): the invitation to delete
        """
        blockchain = cls._get_blockchain(blockchain_id)
        blockchain.delete_invitation(invitation)

    @ classmethod
    def get_walytis_beta_version(cls, ) -> tuple[int, int, int]:
        """Get the software version of the locally running Walytis node.

        Returns:
            tuple: the software version of the locally running Walytis_Beta node
        """
        return WALYTIS_BETA_CORE_VERSION

    @ classmethod
    def get_walytis_beta_api_version(cls, ) -> tuple[int, int, int]:
        """Get the software version of the this walytis_beta_api library.

        Returns:
            tuple: the software version of the this walytis_beta_api library
        """
        return WALYTIS_BETA_API_VERSION

    @ classmethod
    def _get_blockchain(cls, blockchain_id: str) -> walytis_beta.Blockchain:
        blockchain_id = cls.get_blockchain_id(blockchain_id)
        blockchain = walytis_beta.get_blockchain(blockchain_id)
        if blockchain:
            return blockchain
        raise NoSuchBlockchainError(blockchain_id=blockchain_id)

    @ classmethod
    def get_and_read_block(cls, short_id: bytearray) -> Block:
        """Given a block's ID, download and verify it.

        Args:
            short_id (str): the ID of the block to get
        Returns:
            Block: an object representing the specified block, containing all its
                    metadata and block content
        """
        raise Exception(
            "This method hasn't yet been implemented for Walytis Embedded API"
        )

    @ classmethod
    def read_block(cls, block_data: bytearray, ipfs_cid: str) -> Block:
        """Read a block-file, generating a block object if it's valid.

        Args:
            block_data (bytearray): the raw data of the block to read.
            ipfs_cid (str): the IPFS content ID of the block
        Returns:
            Block: a Block object
        """
        raise Exception(
            "This method hasn't yet been implemented for Walytis Embedded API"
        )



"""The fundamental machinery constituting Blocks."""

import hashlib
from copy import deepcopy

# from random import seed
# from random import randint
from datetime import datetime
from datetime import timezone
from secrets import randbits

from walytis_beta_tools.log import logger_block_model as logger
from brenthy_tools_beta.utils import (  # pylint: disable=unused-import
    bytes_from_b255_no_0s,
    bytes_to_b255_no_0s,
    decode_timestamp,
    encode_timestamp,
    from_b255_no_0s,
    to_b255_no_0s,
)
from brenthy_tools_beta.version_utils import decode_version, encode_version

from .exceptions import BlockIntegrityError, InvalidBlockIdError
from ._experimental.generic_block import _GenericBlockImpl as GenericBlock

PREFERRED_HASH_ALGORITHM = "sha512"


class Block(GenericBlock):
    """The Walytis_Beta block.

    This class contains the basic functionality of the Walytis_Beta block,
    used by both Walytis_Beta and walytis_beta_api.
    The networking parts for publishing blocks is in a child class in
    Walytis_Beta.block_networking.
    """

    def __init__(self):
        """Initialise an empty block object, not yet valid on a blockchain."""
        self._ipfs_cid = ""
        self._short_id = bytearray()
        self._long_id = bytearray()
        self._creator_id = bytearray()
        self._creation_time = datetime.now(timezone.utc)
        self._topics = []
        self._content = bytearray()
        self._parents = []
        self._file_data = bytearray()

        self._content_length = 0
        self._content_hash_algorithm = ""
        self._content_hash = bytearray()
        self._n_parents = 0
        self._parents_hash_algorithm = ""
        self._parents_hash = bytearray()
        # the file data which gets published on the blockchain
        self._blockchain_version = []

        self._genesis = False

        self.__integrity_checked = False

    @classmethod
    def from_id(
        cls, long_id: bytearray
    ):
        block = cls()
        block._long_id = long_id
        return block

    @classmethod
    def from_metadata(
        cls,
        ipfs_cid: str,
        creator_id: bytearray,
        creation_time: datetime,
        topics: list[str],
        content_length: int,
        content_hash_algorithm: str,
        content_hash: bytearray,
        content: bytearray,
        n_parents: int,
        parents_hash_algorithm: str,
        parents_hash: bytearray,
        parents: list[bytearray],
        file_data: bytearray,
        blockchain_version: tuple,

    ) -> 'Block':
        """Create a block object with the given fields.

        CAREFUL: No checks or generations are performed.
        """
        block = cls()
        block._ipfs_cid = ipfs_cid
        block._creator_id = creator_id
        block._creation_time = creation_time
        block._topics = topics
        block._content_length = content_length
        block._content_hash_algorithm = content_hash_algorithm
        block._content_hash = content_hash
        block._content = content
        block._n_parents = n_parents
        block._parents_hash_algorithm = parents_hash_algorithm
        block._parents_hash = parents_hash
        block._parents = parents
        block._file_data = file_data
        block._blockchain_version = blockchain_version
        return block

    def generate_content_hash(self) -> bytearray | None:
        """Generate the cryptographic hash of this block's content.

        Returns:
            bytearray: this block's content hash
        """
        self._content_hash_algorithm = PREFERRED_HASH_ALGORITHM
        if not self.content:
            logger.error("Block Hash Generation: empty content")
            return None
        hashGen = hashlib.sha512()
        hashGen.update(self.content)
        content_hash = bytearray.fromhex(hashGen.hexdigest())

        # removing 0s:
        self._content_hash = bytes_to_b255_no_0s(content_hash)
        return self._content_hash

    def generate_parents_hash(self) -> bytearray:
        """Generate the hash of this block's parents' parents-hashes.

        Returns:
            bytearray: this block's parents-hash
        """
        self._parents_hash_algorithm = PREFERRED_HASH_ALGORITHM

        if self._n_parents == 0:  # genesis block
            hashGen = hashlib.sha512()
            hashGen.update(
                self.creator_id
                + encode_timestamp(self.creation_time)
                + to_b255_no_0s(randbits(512))
            )
            parents_hash = bytearray.fromhex(hashGen.hexdigest())
        else:
            # collect a list of parent blocks' parents_hashes
            parent_hashes = bytearray([])
            for parent_id in self.parents:
                parent_hash = decode_short_id(parent_id)["parents_hash"]
                # decode parent hash from its 0-less form
                parent_hash = bytes_from_b255_no_0s(parent_hash)  # returns int

                parent_hashes += parent_hash
            # parents_hash is hash of parent_hashes
            hashGen = hashlib.sha512()
            hashGen.update(parent_hashes)
            parents_hash = bytearray.fromhex(hashGen.hexdigest())

        # removing 0s:
        self._parents_hash = bytes_to_b255_no_0s(parents_hash)
        return self._parents_hash

    def check_integrity(self, ignore_filedata: bool = False) -> bool:
        """Make sure this block's data is complete, self-consistent and valid.

        Args:
            ignore_filedata: if True, checking filedata is skipped, though
                                only if the `file_data` attribute is `None`

        Returns:
            bool: whether this block's data is self-consistent and valid.
        """
        if not self._blockchain_version:
            logger.warning("Block Integrity Check: blockchain_version not set.")
            return False
        if not self.ipfs_cid:
            logger.warning("Block Integrity Check: ipfs_cid not set.")
            return False
        if not self.creator_id:
            logger.warning("Block Integrity Check: creator_id not set.")
            return False
        if not self.creation_time:
            logger.warning("Block Integrity Check: creation_time not set.")
            return False
        if not self.topics:
            logger.warning("Block Integrity Check: topics not set.")
            return False
        if not self._content_length:
            logger.warning("Block Integrity Check: Content length not set.")
            return False
        if not self._content_hash_algorithm:
            logger.warning(
                "Block Integrity Check: content_hash_algorithm not set."
            )
            return False
        if not self._content_hash:
            logger.warning("Block Integrity Check: Hash not set.")
            return False
        # if not self._n_parents:
        #     logger.warning("Block Integrity Check: n_parents not set.")
        #     return None
        if not self._parents_hash_algorithm:
            logger.warning(
                "Block Integrity Check: parents_hash_algorithm not set."
            )
            return False
        if not self._parents_hash:
            logger.warning("Block Integrity Check: parents_hash not set.")
            return False
        if not self.short_id:
            logger.warning("Block Integrity Check: short_id not set.")
            return False
        if not self.long_id:
            logger.warning("Block Integrity Check: long_id not set.")
            return False

        # only skip checking file data if it is not set
        if ignore_filedata and self.file_data is not None:
            ignore_filedata = False

        if not self.file_data and not ignore_filedata:
            logger.warning("Block Integrity Check: filedata not set.")
            return False

        # check if encoded content length matches actual block content length
        if not len(self.content) == self._content_length:
            logger.warning("Block INtegrity Check: content length incorrect")
            return False
        # checking if encoded parents count matches actual parent's count
        if not len(self.parents) == self._n_parents:
            logger.warning("Block Integrity Check: number of parents incorrect")
            return False

        # check if parents are sorted
        sorted_parents = deepcopy(self.parents)
        sorted_parents.sort()
        if self.parents != sorted_parents:
            logger.warning("Block Integrity Check: parents blocks aren't sorted")
            return False

        # Checking generated parts

        # keeping a copy of old generated parts
        old_short_id = deepcopy(self.short_id)
        old_long_id = deepcopy(self.long_id)
        old_content_hash = deepcopy(self._content_hash)
        old_parents_hash = deepcopy(self._parents_hash)
        if not ignore_filedata:
            old_file_data = deepcopy(self.file_data)

        # checking content hash
        self.generate_content_hash()
        # compare the encoded content-hash with the generated content-hash
        if self._content_hash != old_content_hash:
            logger.warning("Block Integrity Check: hash does not match.")
            # restoring old non-matching content_hash
            self._content_hash = old_content_hash
            return False

        if self._n_parents > 0:  # genesis blocks' parents can't be checked
            # checking parents hash
            self.generate_parents_hash()
            # comparing the encoded parents-hash with the generated hash
            if self._parents_hash != old_parents_hash:
                logger.warning(
                    "Block Integrity Check: parents_hash does not match."
                )
                # restoring old non-matching content_hash
                self._parents_hash = old_parents_hash
                return False

            # ensure all parent blocks have older timestamps than this block
            if [
                parent_id
                for parent_id in self.parents
                if (
                    decode_short_id(parent_id)["creation_time"]
                    > self.creation_time
                )
            ]:
                error_message = (
                    "Block Integrity Check: Parent block's creation time "
                    "is greater than this block's creation time "
                    f"{self.creation_time}. This is most likely due to this "
                    "node's clock not being synchronised."
                )
                logger.warning(error_message)
                return False

        # Checking if file_data matches block properties if they are ALL set
        # trying to generate file-data anew from block properties
        self.generate_file_data(check_integrity=False)

        # if the block's file-data and all other properties were set,
        # and the set file-data didn't match all the other propertie,
        # reset the initially set inconsistent file-data property
        if not ignore_filedata and old_file_data != self.file_data:
            self._file_data = old_file_data
            logger.warning(
                "Block Integrity Check: File-data doesn't match block fields."
            )
            return False

        # check if short_id matches block properties
        self.generate_id()
        if old_short_id != self.short_id:
            self._short_id = old_short_id
            logger.warning(
                "Block Integrity Check: short ID doesn't match block fields."
            )
            return False
        if old_long_id != self.long_id:
            self._long_id = old_long_id
            logger.warning(
                "Block Integrity Check: long ID doesn't match block fields."
            )
            return False

        # at this point all integrity tests have been passed
        if not ignore_filedata:
            self.__integrity_checked = True
        return True

    def generate_id(self) -> bytearray | None:
        """Generate this block's ID.

        Returns:
            bytearray: this block's long ID
        """
        # make sure all the necessary components of the short_id have been set
        if not self._blockchain_version:
            logger.warning("Block.generate_id(): blockchain_version not set.")
            return None
        if not self.ipfs_cid:
            logger.warning("Block.generate_id(): ipfs_cid not set.")
            return None
        if not self.creator_id:
            logger.warning("Block.generate_id(): creator_id not set.")
            return None
        if not self.creation_time:
            logger.warning("Block.generate_id(): creation_time not set.")
            return None
        if not self.topics:
            logger.warning("Block.generate_id(): topics not set.")
            return None
        if not self._content_length:
            logger.warning("Block.generate_id(): Content length not set.")
            return None
        if not self._content_hash_algorithm:
            logger.warning("Block.generate_id(): content_hash_algorithm not set.")
            return None
        if not self._content_hash:
            logger.warning("Block.generate_id(): Hash not set.")
            return None
        # if not self._n_parents:
        #     logger.warning("Block.generate_id(): n_parents not set.")
        #     return None
        if not self._parents_hash_algorithm:
            logger.warning("Block.generate_id(): parents_hash_algorithm not set.")
            return None
        if not self._parents_hash:
            logger.warning("Block.generate_id(): parents_hash not set.")
            return None

        self._short_id = bytearray(
            encode_version(self._blockchain_version)
            + bytearray([0, 0])
            + self.ipfs_cid.encode()
            + bytearray([0, 0])
            + self.creator_id
            + bytearray([0, 0])
            + encode_timestamp(self.creation_time)
            + bytearray([0, 0])
            + bytearray([0]).join([topic.encode() for topic in self.topics])
            + bytearray([0, 0])
            + to_b255_no_0s(self._content_length)
            + bytearray([0, 0])
            + self._content_hash_algorithm.encode()
            + bytearray([0, 0])
            + self._content_hash
            + bytearray([0, 0])
            + to_b255_no_0s(self._n_parents)
            + bytearray([0, 0])
            + self._parents_hash_algorithm.encode()
            + bytearray([0, 0])
            + self._parents_hash
        )

        # Generate long ID:

        # add separator between short ID and parents
        long_id = self._short_id + bytearray([0, 0, 0, 0])

        # add parents, separated by [0,0,0]
        long_id += bytearray([0, 0, 0]).join(self.parents)

        self._long_id = long_id

        decode_long_id(long_id)
        return long_id

    def generate_file_data(
        self, check_integrity: bool = False
    ) -> bytearray | None:
        """Generate this block's block file.

        Encodes all the block's data and metadata into a bytearray.

        Args:
            check_integrity (bool): ensure that the block's data is consitent,
                                    raise an exception if not
        Returns:
            bytearray: the block's block-file data
        """
        try:
            data = bytearray(
                encode_version(self._blockchain_version)
                + bytearray([0, 0])
                + self.creator_id
                + bytearray([0, 0])
                + encode_timestamp(self.creation_time)
                + bytearray([0, 0])
                + bytearray([0]).join(
                    [topic.encode() for topic in self.topics]
                )
                + bytearray([0, 0])
                + to_b255_no_0s(self._content_length)
                + bytearray([0, 0])
                + self._content_hash_algorithm.encode()
                + bytearray([0, 0])
                + self._content_hash
                + bytearray([0, 0])
                + to_b255_no_0s(self._n_parents)
                + bytearray([0, 0])
                + self._parents_hash_algorithm.encode()
                + bytearray([0, 0])
                + self._parents_hash
            )
            if self.parents:
                data += bytearray([0, 0, 0, 0]) + bytearray([0, 0, 0]).join(
                    self.parents
                )
            data += bytearray([0, 0, 0, 0, 0]) + self.content

            self._file_data = data

            if check_integrity and not self.check_integrity():
                self._file_data = None
                raise BlockIntegrityError("Integrity check failed")
            return self.file_data
        except Exception as e:
            logger.error("Failed to generate block data.\n" + str(e))
            return None

    def save_block_to_file(self, path: str) -> None:
        """Write this block's block-file-data to a file.

        Args:
            path (str): the filepath to write the blockfile to.
        """
        if len(self.file_data) > 0:
            file = open(path, "wb")
            file.write(self.file_data)
            file.close()
        else:
            logger.error("Couldn't save block to file cause data is null.")


def decode_short_id(short_id: bytearray) -> dict:
    # pylint: disable=trailing-whitespace
    """Extract all the information of a block that is encoded in its ID.

    Args:
        short_id (bytearray): the block ID to decode

    Returns:
        dict: a dictionary containing all the block properties extracted from
            the block ID
        `blockchain_version` (list): Walytis_Beta version of the block
        `ipfs_cid` (str): the IPFS content ID the block was published with
        `creator_id` (bytearray): the ID of the block's creator
        `creation_time` (datetime): the time at which the block was created
        `topics` (list<str>): list of topics which this block belongs to
        `content_length` (int): length of block content in bytes
        `content_hash_algorithm` (str): hash algorithm used for content_hash
        `content_hash` (str): the cryptographic hash of the block's content
        `n_parents` (int): the number of parent blocks which this block has
        `parents_hash_algorithm` (str): hash algorithm used for parents_hash
        `parents_hash` (str): the cryptographic hash of the block's parents
    """
    while short_id[-1] == 0:
        short_id = short_id[:-1]
    metadata = short_id.split(bytearray([0, 0]))
    # logger.important(str(metadata))
    topics = [
        element.decode("utf-8")
        for element in metadata[4].split(bytearray([0]))
    ]
    try:
        return {
            "blockchain_version": decode_version(metadata[0]),
            "ipfs_cid": metadata[1].decode("utf-8"),
            "creator_id": metadata[2],
            "creation_time": decode_timestamp(metadata[3]),
            "topics": topics,
            "content_length": from_b255_no_0s(metadata[5]),
            "content_hash_algorithm": metadata[6].decode(),
            "content_hash": metadata[7],
            "n_parents": from_b255_no_0s(metadata[8]),
            "parents_hash_algorithm": metadata[9].decode(),
            "parents_hash": metadata[10],
        }
    except:
        raise InvalidBlockIdError()


def decode_long_id(long_id: bytearray) -> dict:
    # pylint: disable=trailing-whitespace
    """Extract all the information of a block that is encoded in its ID.

    Args:
        long_id (bytearray): the block ID to decode
    Returns:
        dict: a dictionary containing all the block properties extracted from
            the block ID
        `blockchain_version` (list): Walytis_Beta version of the block
        `ipfs_cid` (str): the IPFS content ID the block was published with
        `creator_id` (bytearray): the ID of the block's creator
        `creation_time` (datetime): the time at which the block was created
        `topics` (list<str>): list of topics which this block belongs to
        `content_length` (int): length of block content in bytes
        `content_hash_algorithm` (str): hash algorithm used for content_hash
        `content_hash` (str): the cryptographic hash of the block's content
        `n_parents` (int): the number of parent blocks which this block has
        `parents_hash_algorithm` (str): hash algorithm used for parents_hash
        `parents_hash` (str): the cryptographic hash of the block's parents
        `parents` (list): list this block's parent blocks
    """
    if not isinstance(long_id, bytearray):
        raise TypeError(
            f"long_id must be of type bytearray, not {type(long_id)}"
        )
    try:
        short_id, parents_data = long_id.split(bytearray([0, 0, 0, 0]))
        result = decode_short_id(short_id)

        parents = parents_data.split(bytearray([0, 0, 0]))
        if parents == [bytearray([])]:  # genesis block
            parents = []
        result.update({"parents": parents})
    except:
        try:
            _ = decode_short_id(long_id)  # noqa
            error_message = "\nYou passed a short ID instead of a long ID."
        except:
            error_message = ""
        raise InvalidBlockIdError(error_message)
    return result


def short_from_long_id(long_id: bytearray) -> bytearray:
    """Convert a long ID into a short ID.

    Args:
        long_id (bytearray): long block ID
    Returns:
        bytearray: the extracted short block ID
    """
    if not (isinstance(long_id, (bytearray, bytes))):
        raise TypeError(f"long_id must be of type bytearray, not {type(long_id)}")
    try:
        short_id = long_id.split(bytearray([0, 0, 0]))[0]
    except:
        raise InvalidBlockIdError()
    return short_id



"""The exceptions used in Walytis_Beta API.

Users of `walytis_beta_api` should use these classes to catch specific errors,
but should not instantiate and raise them themselves, and the full backwards-
compatibility of their constructors will not be guaranteed.
"""


class WalytisReplyDecodeError(Exception):
    """`walytis_beta_api` couldn't decode a reply from Walytis_Beta."""

    def_message = (
        "error parsing the reply from Walytis_Beta. This is probably a bug."
    )

    def __init__(self, message: str = def_message, reply: str = ""):
        self.message = message
        self.reply = str(reply)

    def __str__(self):
        return self.message + " \n" + self.reply


class WalytisBugError(Exception):
    """Walytis_Beta ran into an exception while processing a request."""

    def_message = (
        "Walytis_Beta ran into a problem processing a request. This is a bug."
    )

    def __init__(self, message: str = def_message, error_message: str = ""):
        self.message = message
        self.error_message = message

    def __str__(self):
        if self.error_message:
            return f"{self.message}\nWalytis says: {self.error_message}"
        return self.message


class NoSuchBlockchainError(Exception):
    """The blockchain mentioned by a request does not exist."""

    message = "We have no blockchain with the specified name/id."

    def __init__(self, blockchain_id: str = "", blockchain_name: str = ""):
        self.blockchain_id = blockchain_id
        self.blockchain_name = blockchain_name

    def __str__(self):
        if self.blockchain_id:
            self.message = "We have no blockchain with the specified id: "
            self.message += self.blockchain_id
        if self.blockchain_name:
            self.message = "We have no blockchain with the specified name: "
            self.message += self.blockchain_name
        return self.message


class BlockchainAlreadyExistsError(Exception):
    """We already participate the blockchain trying to be joined."""

    message = "We already have a blockchain with the specified id or name."

    def __init__(self, blockchain_id: str = "", blockchain_name: str = ""):
        self.blockchain_id = blockchain_id
        self.blockchain_name = blockchain_name

    def __str__(self):
        if self.blockchain_id:
            self.message = (
                "We already have a blockchain with the specified id: "
            )
            self.message += self.blockchain_id
        if self.blockchain_name:
            self.message = (
                "We already have a blockchain with the specified name: "
            )
            self.message += self.blockchain_name
        return self.message


class NoSuchInvitationError(Exception):
    """The invitation mentioned by the request does not exist."""

    message = "This blockchain does not have such a join-key."

    def __init__(self, message: str = message):
        self.message = message

    def __str__(self):
        return self.message


class JoinFailureError(Exception):
    """Joining a blockchain failed."""

    def_message = (
        "Failed to join the specified blockchain. Trying again later may work."
    )

    def __init__(self, message: str = def_message, error_message: str = ""):
        self.message = message
        self.error_message = error_message

    def __str__(self):
        if self.error_message:
            return self.message + "\n" + self.error_message
        return self.message


class BlockNotFoundError(Exception):
    """The block mentioned by the request doesn't exists/isn't known yet."""

    def_message = (
        "Block not found!"
    )

    def __init__(self, message: str = def_message):
        self.message = message

    def __str__(self):
        return self.message


class BlockCreationError(Exception):
    """The walytis_beta blockchain failed to add a block to the blockchain."""

    def_message = "Failed to add a block to a blockchain."

    def __init__(self, message: str = def_message, blockchain_name: str = ""):
        self.message = message
        self.blockchain_name = blockchain_name

    def __str__(self):
        if self.blockchain_name:
            return f"{self.message} Blockchain: {self.blockchain_name}"
        return self.message


class InvalidBlockIdError(Exception):
    """When a variable isn't a valid block ID."""

    def_message = "The given data is not a valid block ID."

    def __init__(
        self, message: str = ""
    ):
        if message:
            self.message = message
        else:
            self.message = self.def_message

    def __str__(self):
        return self.message


class BlockIntegrityError(Exception):
    """A block object doesn't satisfy all the Walytis_Beta block constraints.

    This is due to a software bug or a failed forgery attempt.
    """

    def __init__(
        self, message: str, sought_id: str = "", received_id: str = ""
    ):
        self.message = message
        self.sought_id = sought_id
        self.received_id = received_id

    def __str__(self):
        stri = self.message
        if self.sought_id:
            stri += "\n" + str(self.sought_id)
        if self.received_id:
            stri += "\n" + str(self.received_id)
        return stri


class NotSupposedToHappenError(Exception):
    """When a should-be-impossible situation arises, indicating a bug."""

    def_message = "Something weird happened. This is a bug."

    def __init__(self, message: str = def_message):
        self.message = message

    def __str__(self):
        return self.message


NO_SUCH_BLOCKCHAIN_MESSAGE = "no such blockchain"
NO_SUCH_INVITATION_MESSAGE = "no such join-key"
BLOCK_NOT_FOUND = "block not found"
WALYTIS_BETA_ERROR_MESSAGE = "internal Walytis_Beta error"
BLOCKCHAIN_EXISTS_MESSAGE = "blockchain already exists"


import os
from ipfs_tk_generics.client import IpfsClient
from walytis_beta_tools.log import logger

from enum import Enum
from environs import Env

env = Env()
# initialise IPFS
USE_IPFS_NODE = os.environ.get("USE_IPFS_NODE", "").lower() in ["true", "1"]
DEF_IPFS_REPO_DIR = os.path.abspath(os.path.join(".ipfs_repo"))
IPFS_REPO_DIR = env.str("IPFS_REPO_DIR", DEF_IPFS_REPO_DIR)
ipfs: IpfsClient


class IpfsTkModes(Enum):
    EMBEDDED = "EMBEDDED"
    HTTP = "HTTP"


def get_ipfs_tk_mode() -> IpfsTkModes:
    return env.enum(
        "IPFS_TK_MODE",
        enum=IpfsTkModes,
        default=IpfsTkModes.HTTP,
    )


class WalytisTestModes(Enum):
    EMBEDDED = "EMBEDDED"
    USE_BRENTHY = "USE_BRENTHY"
    RUN_BRENTHY = "RUN_BRENTHY"


def get_walytis_test_mode() -> WalytisTestModes:
    return env.enum(
        "WALYTIS_TEST_MODE",
        enum=WalytisTestModes,
        default=WalytisTestModes.EMBEDDED,
    )


class WalytisBetaApiTypes(Enum):
    WALYTIS_BETA_BRENTHY_API = 0
    WALYTIS_BETA_DIRECT_API = 1


def get_walytis_beta_api_type():
    return env.enum(
        "WALYTIS_BETA_API_TYPE",
        enum=WalytisBetaApiTypes,
        default=WalytisBetaApiTypes.WALYTIS_BETA_BRENTHY_API,
    )
